from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsPathItem.__bases__ = (widgets.AbstractGraphicsShapeItem,)


class GraphicsPathItem(QtWidgets.QGraphicsPathItem):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QAbstractListModel.__bases__ = (core.AbstractItemModel,)


class AbstractListModel(QtCore.QAbstractListModel):
    pass



from __future__ import annotations

from typing import Union

from prettyqt import core, gui
from prettyqt.qt import QtGui


QtGui.QRegularExpressionValidator.__bases__ = (gui.Validator,)


class RegularExpressionValidator(QtGui.QRegularExpressionValidator):
    def __repr__(self):
        return f"{type(self).__name__}(RegularExpression({self.get_regex()!r}))"

    def __getstate__(self):
        return dict(pattern=core.RegularExpression(self.regularExpression()))

    def __setstate__(self, state):
        self.setRegularExpression(state["pattern"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __eq__(self, other: object):
        if not isinstance(other, type(self)):
            return False
        return self.regularExpression() == other.regularExpression()

    def set_regex(self, regex: Union[str, core.RegularExpression], flags=0):
        if isinstance(regex, str):
            regex = core.RegularExpression(regex, flags)
        self.setRegularExpression(regex)

    def get_regex(self) -> str:
        val = self.regularExpression()
        return val.pattern()


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    w = widgets.LineEdit()
    val = RegularExpressionValidator()
    val.set_regex(r"\w\d\d")
    w.set_validator(val)
    w.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import core, gui
from prettyqt.qt import QtGui


QtGui.QRadialGradient.__bases__ = (gui.Gradient,)


class RadialGradient(QtGui.QRadialGradient):
    def __repr__(self):
        return (
            f"{type(self).__name__}({self.get_center()}, {self.centerRadius()}, "
            f"{self.get_focal_point()}, {self.focalRadius()})"
        )

    def serialize_fields(self):
        center = self.center()
        focal_point = self.focalPoint()
        return dict(
            center_radius=self.centerRadius(),
            radius=self.radius(),
            focal_radius=self.focalRadius(),
            center=(center[0], center[1]),
            focal_point=(focal_point[0], focal_point[1]),
        )

    def get_center(self) -> core.PointF:
        return core.PointF(self.center())

    def get_focal_point(self) -> core.PointF:
        return core.PointF(self.focalPoint())



from __future__ import annotations

from prettyqt import charts
from prettyqt.qt.QtCharts import QtCharts


QtCharts.QStackedBarSeries.__bases__ = (charts.AbstractBarSeries,)


class StackedBarSeries(QtCharts.QStackedBarSeries):
    pass



from __future__ import annotations

from prettyqt import charts
from prettyqt.qt.QtCharts import QtCharts


QtCharts.QHorizontalPercentBarSeries.__bases__ = (charts.AbstractBarSeries,)


class HorizontalPercentBarSeries(QtCharts.QHorizontalPercentBarSeries):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsEffect.__bases__ = (core.Object,)


class GraphicsEffect(QtWidgets.QGraphicsEffect):
    def serialize_fields(self):
        return dict(enabled=self.isEnabled())

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setEnabled(state["enabled"])



from __future__ import annotations

from prettyqt.qt import QtPositioning
from prettyqt.utils import bidict


SHAPE_TYPES = bidict(
    unknown=QtPositioning.QGeoShape.UnknownType,
    rectangle=QtPositioning.QGeoShape.RectangleType,
    circle=QtPositioning.QGeoShape.CircleType,
    path=QtPositioning.QGeoShape.PathType,
    polygon=QtPositioning.QGeoShape.PolygonType,
)


class GeoShape(QtPositioning.QGeoShape):
    def __contains__(self, other: QtPositioning.QGeoCoordinate):
        return self.contains(other)

    def __str__(self):
        return self.toString()[1:]

    def get_type(self) -> str:
        return SHAPE_TYPES.inverse[self.type()]



from __future__ import annotations

import os
from typing import Iterator, Literal, Union

from prettyqt import constants, core, gui
from prettyqt.qt import QtGui
from prettyqt.utils import InvalidParamError, bidict


if core.VersionNumber.get_qt_version() >= (5, 14, 0):
    MARKDOWN_FEATURES = bidict(
        no_html=QtGui.QTextDocument.MarkdownNoHTML,
        commonmark=QtGui.QTextDocument.MarkdownDialectCommonMark,
        github=QtGui.QTextDocument.MarkdownDialectGitHub,
    )

MarkdownFeatureStr = Literal["no_html", "commonmark", "github"]

RESOURCE_TYPES = bidict(
    unknown=QtGui.QTextDocument.UnknownResource,
    html=QtGui.QTextDocument.HtmlResource,
    image=QtGui.QTextDocument.ImageResource,
    stylesheet=QtGui.QTextDocument.StyleSheetResource,
    markdown=QtGui.QTextDocument.MarkdownResource,
    user=QtGui.QTextDocument.UserResource,
)

ResourceTypeStr = Literal["unknown", "html", "image", "stylesheet", "markdown", "user"]

STACKS = bidict(
    undo=QtGui.QTextDocument.UndoStack,
    redo=QtGui.QTextDocument.RedoStack,
    undo_and_redo=QtGui.QTextDocument.UndoAndRedoStacks,
)

StackStr = Literal["undo", "redo", "undo_and_redo"]

FIND_FLAGS = bidict(
    backward=QtGui.QTextDocument.FindBackward,
    case_sensitive=QtGui.QTextDocument.FindCaseSensitively,
    whole_words=QtGui.QTextDocument.FindWholeWords,
)

FindFlagStr = Literal["backward", "case_sensitive", "whole_words"]

QtGui.QTextDocument.__bases__ = (core.Object,)


class TextDocument(QtGui.QTextDocument):
    def __getitem__(self, index: int) -> gui.TextBlock:
        return gui.TextBlock(self.findBlockByNumber(index))

    def __len__(self) -> int:
        return self.blockCount()

    def __iter__(self) -> Iterator[gui.TextBlock]:
        return iter(self[i] for i in range(self.blockCount()))

    def set_text(self, text: str):
        self.setPlainText(text)

    def serialize_fields(self):
        return dict(
            base_url=self.get_base_url(),
            default_font=self.get_default_font(),
            default_stylesheet=self.defaultStyleSheet(),
            default_text_option=self.get_default_text_option(),
            document_margin=self.documentMargin(),
            maximum_block_count=self.maximumBlockCount(),
            is_modified=self.isModified(),
            page_size=self.pageSize(),
            text_width=self.textWidth(),
            indent_width=self.indentWidth(),
            undo_redo_enabled=self.isUndoRedoEnabled(),
            use_design_metrics=self.useDesignMetrics(),
        )

    def get_base_url(self) -> core.Url:
        return core.Url(self.baseUrl())

    def get_default_font(self) -> gui.Font:
        return gui.Font(self.defaultFont())

    def get_default_text_option(self) -> gui.TextOption:
        return gui.TextOption(self.defaultTextOption())

    def clear_stacks(self, stack: StackStr):
        """Clear undo / redo stack.

        Args:
            stack: stack to clear

        Raises:
            InvalidParamError: stack type does not exist
        """
        if stack not in STACKS:
            raise InvalidParamError(stack, STACKS)
        self.clearUndoRedoStacks(STACKS[stack])

    def set_default_cursor_move_style(self, style: constants.CursorMoveStyleStr):
        """Set the cursor move style.

        Args:
            style: cursor move style

        Raises:
            InvalidParamError: cursor move style does not exist
        """
        if style not in constants.CURSOR_MOVE_STYLE:
            raise InvalidParamError(style, constants.CURSOR_MOVE_STYLE)
        self.setDefaultCursorMoveStyle(constants.CURSOR_MOVE_STYLE[style])

    def get_default_cursor_move_style(self) -> constants.CursorMoveStyleStr:
        """Return current cursor move style.

        Returns:
            cursor move style
        """
        return constants.CURSOR_MOVE_STYLE.inverse[self.defaultCursorMoveStyle()]

    def add_resource(
        self, resource_type: ResourceTypeStr, name: Union[str, os.PathLike], resource
    ):
        if resource_type not in RESOURCE_TYPES:
            raise InvalidParamError(resource_type, RESOURCE_TYPES)
        url = core.Url(name)
        self.addResource(RESOURCE_TYPES[resource_type], url, resource)


if __name__ == "__main__":
    doc = TextDocument("This is a test\nHello")
    for i in doc:
        print(i)



"""Custom_models module.

Contains custom models
"""

from prettyqt import core
from .selectionmixin import SelectionMixin

if core.VersionNumber.get_qt_version() < (6, 0, 0):
    from .playlistmodel import PlaylistModel
from .transposeproxymodel import TransposeProxyModel
from .importlibdistributionmodel import ImportlibDistributionModel
from .regexmatchesmodel import RegexMatchesModel
from .columnitemmodel import ColumnItemModel, ColumnItem

__all__ = [
    "SelectionMixin",
    "TransposeProxyModel",
    "ImportlibDistributionModel",
    "PlaylistModel",
    "RegexMatchesModel",
    "ColumnItemModel",
    "ColumnItem",
]



from __future__ import annotations

from prettyqt import core, widgets
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsRectItem.__bases__ = (widgets.AbstractGraphicsShapeItem,)


class GraphicsRectItem(QtWidgets.QGraphicsRectItem):
    def __repr__(self):
        return f"{type(self).__name__}({repr(self.get_rect())})"

    def get_rect(self):
        return core.RectF(self.rect())

    def serialize_fields(self):
        return dict(rect=self.get_rect())

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setRect(state["rect"])



from __future__ import annotations

import contextlib
from typing import Iterator, Literal, Optional

from deprecated import deprecated

from prettyqt import constants, core, gui, syntaxhighlighters, widgets
from prettyqt.qt import QtGui, QtWidgets
from prettyqt.utils import InvalidParamError, bidict


LINE_WRAP_MODE = bidict(
    none=QtWidgets.QPlainTextEdit.NoWrap,
    widget_width=QtWidgets.QPlainTextEdit.WidgetWidth,
)

LineWrapModeStr = Literal["none", "widget_width"]


QtWidgets.QPlainTextEdit.__bases__ = (widgets.AbstractScrollArea,)


class PlainTextEdit(QtWidgets.QPlainTextEdit):

    value_changed = core.Signal()

    def __init__(
        self,
        text: str = "",
        parent: Optional[QtWidgets.QWidget] = None,
        read_only: bool = False,
    ):
        super().__init__(text, parent)
        self._allow_wheel_zoom = False
        self.validator: Optional[QtGui.QValidator] = None
        self.textChanged.connect(self._on_value_change)
        self.set_read_only(read_only)

    def serialize_fields(self):
        return dict(
            text=self.text(),
            read_only=self.isReadOnly(),
            line_wrap_mode=self.get_line_wrap_mode(),
            word_wrap_mode=self.get_word_wrap_mode(),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self.set_text(state["text"])
        self.setReadOnly(state["read_only"])
        self.set_line_wrap_mode(state["line_wrap_mode"])
        self.set_word_wrap_mode(state["word_wrap_mode"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __add__(self, other: str):
        self.append_text(other)
        return self

    def wheelEvent(self, event):
        """Handle wheel event for zooming."""
        if not self._allow_wheel_zoom:
            return None
        if event.modifiers() & constants.CTRL_MOD:
            self.zoomIn() if event.angleDelta().y() > 0 else self.zoomOut()
        else:
            super().wheelEvent(event)

    def allow_wheel_zoom(self, do_zoom: bool = True):
        self._allow_wheel_zoom = do_zoom

    @contextlib.contextmanager
    def create_cursor(self) -> Iterator[gui.TextCursor]:
        cursor = gui.TextCursor(self.document())
        yield cursor
        self.setTextCursor(cursor)

    @contextlib.contextmanager
    def current_cursor(self) -> Iterator[gui.TextCursor]:
        cursor = gui.TextCursor(self.textCursor())
        yield cursor
        self.setTextCursor(cursor)

    def move_cursor(
        self,
        operation: gui.textcursor.MoveOperationStr,
        mode: gui.textcursor.MoveModeStr = "move",
    ):
        self.moveCursor(
            gui.textcursor.MOVE_OPERATION[operation], gui.textcursor.MOVE_MODE[mode]
        )

    def append_text(self, text: str, newline: bool = True):
        if newline:
            self.appendPlainText(text)
        else:
            self.move_cursor("end")
            self.insertPlainText(text)
            self.move_cursor("end")

    def set_text(self, text: str):
        self.setPlainText(text)

    def set_syntaxhighlighter(self, syntax: str, style: Optional[str] = None):
        self._hl = syntaxhighlighters.PygmentsHighlighter(self.document(), syntax)
        if style is not None:
            self._hl.set_style(style)

    def text(self) -> str:
        return self.toPlainText()

    def select_text(self, start: int, end: int):
        with self.create_cursor() as c:
            c.select_text(start, end)

    def set_read_only(self, value: bool = True):
        """Make the PlainTextEdit read-only.

        Args:
            value: True, for read-only, otherwise False
        """
        self.setReadOnly(value)

    def highlight_current_line(self, color="yellow"):
        extra_selections = []

        if not self.isReadOnly():
            selection = widgets.TextEdit.ExtraSelection()
            line_color = gui.Color(color)
            selection.format.setBackground(line_color)
            selection.format.setProperty(QtGui.QTextFormat.FullWidthSelection, True)
            selection.cursor = self.textCursor()
            selection.cursor.clearSelection()
            extra_selections.append(selection)

        self.setExtraSelections(extra_selections)

    @deprecated(reason="This method is deprecated, use set_word_wrap_mode instead.")
    def set_wrap_mode(self, mode: gui.textoption.WordWrapModeStr):
        self.set_word_wrap_mode(mode)

    def set_word_wrap_mode(self, mode: gui.textoption.WordWrapModeStr):
        """Set word wrap mode.

        Args:
            mode: word wrap mode to use

        Raises:
            InvalidParamError: wrap mode does not exist
        """
        if mode not in gui.textoption.WORD_WRAP_MODE:
            raise InvalidParamError(mode, gui.textoption.WORD_WRAP_MODE)
        self.setWordWrapMode(gui.textoption.WORD_WRAP_MODE[mode])

    @deprecated(reason="This method is deprecated, use get_word_wrap_mode instead.")
    def get_wrap_mode(self) -> gui.textoption.WordWrapModeStr:
        return self.get_word_wrap_mode()

    def get_word_wrap_mode(self) -> gui.textoption.WordWrapModeStr:
        """Get the current word wrap mode.

        Returns:
            Word wrap mode
        """
        return gui.textoption.WORD_WRAP_MODE.inverse[self.wordWrapMode()]

    def set_line_wrap_mode(self, mode: LineWrapModeStr):
        """Set line wrap mode.

        Args:
            mode: line wrap mode to use

        Raises:
            InvalidParamError: line wrap mode does not exist
        """
        if mode not in LINE_WRAP_MODE:
            raise InvalidParamError(mode, LINE_WRAP_MODE)
        self.setLineWrapMode(LINE_WRAP_MODE[mode])

    def get_line_wrap_mode(self) -> LineWrapModeStr:
        """Get the current wrap mode.

        Returns:
            Wrap mode
        """
        return LINE_WRAP_MODE.inverse[self.lineWrapMode()]

    def _on_value_change(self):
        self.value_changed.emit()
        if self.validator is not None:
            self._set_validation_color()

    def _set_validation_color(self, state: bool = True):
        color = "orange" if not self.is_valid() else None
        self.set_background_color(color)

    def set_validator(self, validator: Optional[QtGui.QValidator]):
        self.validator = validator
        self._set_validation_color()

    def set_regex_validator(self, regex: str, flags=0) -> gui.RegularExpressionValidator:
        validator = gui.RegularExpressionValidator(self)
        validator.set_regex(regex, flags)
        self.set_validator(validator)
        return validator

    def is_valid(self):
        if self.validator is None:
            return True
        return self.validator.is_valid_value(self.text())

    def set_value(self, value: str):
        self.setPlainText(value)

    def get_value(self):
        return self.text()


if __name__ == "__main__":
    from prettyqt import custom_validators

    val = custom_validators.RegexPatternValidator()
    app = widgets.app()
    widget = PlainTextEdit("This is a test")
    widget.set_validator(val)
    with widget.current_cursor() as c:
        c.select_text(2, 4)
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import bidict


POSITION = bidict(
    left=QtWidgets.QStyleOptionViewItem.Left,
    right=QtWidgets.QStyleOptionViewItem.Right,
    top=QtWidgets.QStyleOptionViewItem.Top,
    bottom=QtWidgets.QStyleOptionViewItem.Bottom,
)

VIEW_ITEM_FEATURE = bidict(
    none=QtWidgets.QStyleOptionViewItem.ViewItemFeature(),
    wrap_text=QtWidgets.QStyleOptionViewItem.WrapText,
    alternate=QtWidgets.QStyleOptionViewItem.Alternate,
    has_check_indicator=QtWidgets.QStyleOptionViewItem.HasCheckIndicator,
    has_display=QtWidgets.QStyleOptionViewItem.HasDisplay,
    has_decoration=QtWidgets.QStyleOptionViewItem.HasDecoration,
)

VIEW_ITEM_POSITION = bidict(
    invalid=QtWidgets.QStyleOptionViewItem.Invalid,
    beginning=QtWidgets.QStyleOptionViewItem.Beginning,
    middle=QtWidgets.QStyleOptionViewItem.Middle,
    end=QtWidgets.QStyleOptionViewItem.End,
    only_one=QtWidgets.QStyleOptionViewItem.OnlyOne,
)

QtWidgets.QStyleOptionViewItem.__bases__ = (widgets.StyleOption,)


class StyleOptionViewItem(QtWidgets.QStyleOptionViewItem):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtQuick


QtQuick.QQuickRenderControl.__bases__ = (core.Object,)


class QuickRenderControl(QtQuick.QQuickRenderControl):
    pass


if __name__ == "__main__":
    item = QuickRenderControl()



from __future__ import annotations

from typing import Tuple

from prettyqt import gui
from prettyqt.qt import QtGui


class NotZeroValidator(gui.Validator):
    def validate(  # type: ignore
        self, text: str, pos: int = 0
    ) -> Tuple[QtGui.QValidator.State, str, int]:
        if text == "0":
            return self.Intermediate, text, pos
        return self.Acceptable, text, pos

    def __eq__(self, other: object):
        return isinstance(other, NotZeroValidator)


if __name__ == "__main__":
    from prettyqt import widgets

    val = NotZeroValidator()
    app = widgets.app()
    widget = widgets.LineEdit("This is a test")
    widget.setValidator(val)
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import constants
from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError


class DeadlineTimer(QtCore.QDeadlineTimer):
    def set_type(self, typ: constants.TimerTypeStr):
        """Set the timer type.

        Args:
            typ: timer type

        Raises:
            InvalidParamError: timer type does not exist
        """
        if typ not in constants.TIMER_TYPE:
            raise InvalidParamError(typ, constants.TIMER_TYPE)
        self.setTimerType(constants.TIMER_TYPE[typ])

    def get_type(self) -> constants.TimerTypeStr:
        """Return current timer type.

        Returns:
            timer type
        """
        return constants.TIMER_TYPE.inverse[self.timerType()]



from __future__ import annotations

from typing import Iterator, List, Union

from prettyqt import core, winextras
from prettyqt.qt import QtWinExtras


QtWinExtras.QWinJumpList.__bases__ = (core.Object,)


class WinJumpList(QtWinExtras.QWinJumpList):
    def __setitem__(
        self,
        key: str,
        val: Union[List[QtWinExtras.QWinJumpListItem], QtWinExtras.QWinJumpListCategory],
    ):
        if isinstance(val, QtWinExtras.QWinJumpListCategory):
            self.addCategory(val)
        else:
            cat = self.add_category(key)
            for item in val:
                cat.addItem(item)

    def __getitem__(self, key: str) -> QtWinExtras.QWinJumpListCategory:
        for i in self.categories():
            if i.title() == key:
                return i
        raise KeyError(key)

    def __iter__(self) -> Iterator[QtWinExtras.QWinJumpListCategory]:
        return iter(self.categories())

    def add_category(self, title: str) -> winextras.WinJumpListCategory:
        cat = winextras.WinJumpListCategory(title)
        self.addCategory(cat)
        return cat

    def get_recent(self) -> QtWinExtras.QWinJumpListCategory:
        return self.recent()

    def get_frequent(self) -> QtWinExtras.QWinJumpListCategory:
        return self.frequent()

    def get_tasks(self) -> QtWinExtras.QWinJumpListCategory:
        return self.tasks()


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    pass
    app.main_loop()



"""Provides QtCore classes and functions."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt.qt import PYQT5, PYSIDE2, PYSIDE6, PythonQtError

if TYPE_CHECKING:
    import datetime


if PYQT5:
    from PyQt5.QtCore import (
        QT_VERSION_STR,
        QT_VERSION_STR as __version__,
        QDateTime,
        QDate,
        QTime,
        pyqtBoundSignal,
        pyqtBoundSignal as SignalInstance,
        pyqtProperty,
        pyqtProperty as Property,
        pyqtSignal,
        pyqtSignal as Signal,
        pyqtSlot,
        pyqtSlot as Slot,
        Q_ENUM,
        Q_ENUM as QEnum,
        Q_FLAG,
        Q_FLAG as QFlag,
    )

    # For issue #153
    from PyQt5.QtCore import *

    def to_datetime(self) -> datetime.datetime:
        return self.toPyDateTime()

    QDateTime.toPython = to_datetime

    def to_date(self) -> datetime.date:
        return self.toPyDate()

    QDate.toPython = to_date

    def to_time(self) -> datetime.time:
        return self.toPyTime()

    QTime.toPython = to_time

    # Those are imported from `import *`
    del (
        pyqtSignal,
        pyqtBoundSignal,
        pyqtSlot,
        pyqtProperty,
        QT_VERSION_STR,
        Q_FLAG,
        Q_ENUM,
    )
elif PYSIDE2:
    from PySide2.QtCore import *
    from PySide2.QtCore import __version__  # type: ignore
elif PYSIDE6:
    from PySide6.QtCore import *  # type: ignore
    from PySide6.QtCore import __version__  # type: ignore



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsTextItem.__bases__ = (widgets.GraphicsObject,)


class GraphicsTextItem(QtWidgets.QGraphicsTextItem):
    def __repr__(self):
        return f"{type(self).__name__}({self.toPlainText()!r})"

    def serialize_fields(self):
        return dict(
            open_external_links=self.openExternalLinks(),
            text_width=self.textWidth(),
            plain_text=self.toPlainText(),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setOpenExternalLinks(state["open_external_links"])
        self.setTextWidth(state["text_width"])
        self.setPlainText(state["plain_text"])



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStyleOptionToolButton.__bases__ = (widgets.StyleOptionComplex,)


class StyleOptionToolButton(QtWidgets.QStyleOptionToolButton):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWidgets


class FadeInAnimation(core.PropertyAnimation):
    def __init__(
        self, duration: int = 1000, easing: core.easingcurve.TypeStr = "in_out_sine"
    ):
        super().__init__()
        self.set_easing(easing)
        self.setStartValue(0.0)
        self.setEndValue(1.0)
        self.setDuration(duration)

    def apply_to(self, obj: QtWidgets.QWidget, attribute: str = "windowOpacity"):
        super().apply_to(obj, attribute)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    val = FadeInAnimation()



from __future__ import annotations

import os
import pathlib
from typing import Optional, Set, Union

from prettyqt import core
from prettyqt.qt import QtCore


LOCALIZATION_PATH = pathlib.Path(__file__).parent.parent / "localization"


QtCore.QTranslator.__bases__ = (core.Object,)


class Translator(QtCore.QTranslator):
    def __bool__(self):
        return not self.isEmpty()

    def get_file_path(self) -> Optional[pathlib.Path]:
        path = self.filePath()
        if not path:
            return None
        return pathlib.Path(path)

    def load_file(self, path: Union[str, os.PathLike]):
        path = pathlib.Path(path)
        if not self.load(path.name, str(path.parent)):
            raise OSError(f"Invalid language file {path}")

    @classmethod
    def get_available_languages(cls) -> Set[str]:
        return {
            str(path).split("_", maxsplit=1)[1][:-3]
            for path in LOCALIZATION_PATH.iterdir()
        }

    @classmethod
    def for_language(cls, language: str) -> Translator:
        # if language not in cls.get_available_languages():
        #     raise ValueError("Language does not exist")
        file = LOCALIZATION_PATH / f"language_{language}.qm"
        translator = cls()
        translator.load_file(file)
        return translator


if __name__ == "__main__":
    print(Translator.get_available_languages())
    app = core.app()
    translator = Translator.for_language("de")



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWinExtras


QtWinExtras.QWinThumbnailToolButton.__bases__ = (core.Object,)


class WinThumbnailToolButton(QtWinExtras.QWinThumbnailToolButton):
    pass


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    btn = WinThumbnailToolButton()
    # app.main_loop()



from __future__ import annotations

from typing import Any, Dict, List, Literal, Tuple

from prettyqt import gui
from prettyqt.qt import QtGui
from prettyqt.utils import InvalidParamError, bidict, prettyprinter


COORDINATE_MODE = bidict(
    logical=QtGui.QGradient.LogicalMode,
    object=QtGui.QGradient.ObjectMode,
    stretch_to_device=QtGui.QGradient.StretchToDeviceMode,
    object_bounding=QtGui.QGradient.ObjectBoundingMode,
)

CoordinateModeStr = Literal["logical", "object", "stretch_to_device", "object_bounding"]

SPREAD = bidict(
    pad=QtGui.QGradient.PadSpread,
    repeat=QtGui.QGradient.RepeatSpread,
    reflect=QtGui.QGradient.ReflectSpread,
)

SpreadStr = Literal["pad", "repeat", "reflect"]

TYPE = bidict(
    linear=QtGui.QGradient.LinearGradient,
    radial=QtGui.QGradient.RadialGradient,
    conical=QtGui.QGradient.ConicalGradient,
    none=QtGui.QGradient.NoGradient,
)

TypeStr = Literal["linear", "radial", "conical", "none"]


PRESET = bidict(
    warm_flame=QtGui.QGradient.WarmFlame,
    night_fade=QtGui.QGradient.NightFade,
    spring_warmth=QtGui.QGradient.SpringWarmth,
    juicy_peach=QtGui.QGradient.JuicyPeach,
    young_passion=QtGui.QGradient.YoungPassion,
    lady_lips=QtGui.QGradient.LadyLips,
    sunny_morning=QtGui.QGradient.SunnyMorning,
    rainy_ashville=QtGui.QGradient.RainyAshville,
    frozen_dreams=QtGui.QGradient.FrozenDreams,
    winter_neva=QtGui.QGradient.WinterNeva,
    dusty_grass=QtGui.QGradient.DustyGrass,
    tempting_azure=QtGui.QGradient.TemptingAzure,
    heavy_rain=QtGui.QGradient.HeavyRain,
    amy_crisp=QtGui.QGradient.AmyCrisp,
    mean_fruit=QtGui.QGradient.MeanFruit,
    deep_blue=QtGui.QGradient.DeepBlue,
    ripe_malinka=QtGui.QGradient.RipeMalinka,
    cloudy_knoxville=QtGui.QGradient.CloudyKnoxville,
    malibu_beach=QtGui.QGradient.MalibuBeach,
    new_life=QtGui.QGradient.NewLife,
    true_sunset=QtGui.QGradient.TrueSunset,
    morpheus_den=QtGui.QGradient.MorpheusDen,
    rare_wind=QtGui.QGradient.RareWind,
    near_moon=QtGui.QGradient.NearMoon,
    wild_apple=QtGui.QGradient.WildApple,
    saint_petersburg=QtGui.QGradient.SaintPetersburg,
    plum_plate=QtGui.QGradient.PlumPlate,
    everlasting_sky=QtGui.QGradient.EverlastingSky,
    happy_fisher=QtGui.QGradient.HappyFisher,
    blessing=QtGui.QGradient.Blessing,
    sharpeye_eagle=QtGui.QGradient.SharpeyeEagle,
    ladoga_bottom=QtGui.QGradient.LadogaBottom,
    lemon_gate=QtGui.QGradient.LemonGate,
    itmeo_branding=QtGui.QGradient.ItmeoBranding,
    zeus_miracle=QtGui.QGradient.ZeusMiracle,
    old_hat=QtGui.QGradient.OldHat,
    star_wine=QtGui.QGradient.StarWine,
    happy_acid=QtGui.QGradient.HappyAcid,
    awesome_pine=QtGui.QGradient.AwesomePine,
    new_york=QtGui.QGradient.NewYork,
    shy_rainbow=QtGui.QGradient.ShyRainbow,
    mixed_hopes=QtGui.QGradient.MixedHopes,
    fly_high=QtGui.QGradient.FlyHigh,
    # ...
)


class Gradient(prettyprinter.PrettyPrinter, QtGui.QGradient):
    def __setitem__(self, key: float, value):
        self.setColorAt(key, value)

    def serialize_fields(self):
        return dict(
            coordinate_mode=self.get_coordinate_mode(),
            spread=self.get_spread(),
            stops=self.get_stops(),
        )

    def serialize(self) -> Dict[str, Any]:
        return self.serialize_fields()

    def set_coordinate_mode(self, mode: CoordinateModeStr):
        """Set the coordinate mode.

        Args:
            mode: coordinate mode

        Raises:
            InvalidParamError: mode does not exist
        """
        if mode not in COORDINATE_MODE:
            raise InvalidParamError(mode, COORDINATE_MODE)
        self.setCoordinateMode(COORDINATE_MODE[mode])

    def get_coordinate_mode(self) -> CoordinateModeStr:
        """Return current coordinate mode.

        Returns:
            coordinate mode
        """
        return COORDINATE_MODE.inverse[self.coordinateMode()]

    def set_spread(self, method: SpreadStr):
        """Set the spread method.

        Args:
            method: spread method

        Raises:
            InvalidParamError: method does not exist
        """
        if method not in SPREAD:
            raise InvalidParamError(method, SPREAD)
        self.setSpread(SPREAD[method])

    def get_spread(self) -> SpreadStr:
        """Return current spread method.

        Returns:
            spread method
        """
        return SPREAD.inverse[self.spread()]

    def get_type(self) -> TypeStr:
        """Return current gradient type.

        Returns:
            gradient type
        """
        return TYPE.inverse[self.type()]

    def get_stops(self) -> List[Tuple[float, gui.Color]]:
        return [(i, gui.Color(j)) for (i, j) in self.stops()]


if __name__ == "__main__":
    grad = Gradient()
    grad.setStops([(0.0, gui.Color("red")), (1.0, gui.Color("green"))])
    print(grad.get_stops())
    print(repr(grad))



from __future__ import annotations

from typing import Iterator, Literal, Union

from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError, bidict


FIELD_ALIGNMENT = bidict(
    left=QtCore.QTextStream.AlignLeft,
    right=QtCore.QTextStream.AlignRight,
    center=QtCore.QTextStream.AlignCenter,
    accounting_style=QtCore.QTextStream.AlignAccountingStyle,
)

FieldAlignmentStr = Literal["left", "right", "center", "accounting_style"]

NUMBER_FLAGS = bidict(
    show_base=QtCore.QTextStream.ShowBase,
    force_point=QtCore.QTextStream.ForcePoint,
    force_sign=QtCore.QTextStream.ForceSign,
    uppercase_base=QtCore.QTextStream.UppercaseBase,
    uppercase_digits=QtCore.QTextStream.UppercaseDigits,
)

NumberFlagStr = Literal[
    "show_base", "force_point", "force_sign", "uppercase_base", "uppercase_digits"
]

REAL_NUMBER_NOTATION = bidict(
    scientific=QtCore.QTextStream.ScientificNotation,
    fixed=QtCore.QTextStream.FixedNotation,
    smart=QtCore.QTextStream.SmartNotation,
)

RealNumberNotationStr = Literal["scientific", "fixed", "smart"]

STATUS = bidict(
    ok=QtCore.QTextStream.Ok,
    read_past_end=QtCore.QTextStream.ReadPastEnd,
    read_corrupt_data=QtCore.QTextStream.ReadCorruptData,
    write_failed=QtCore.QTextStream.WriteFailed,
)

StatusStr = Literal["ok", "read_past_end", "read_corrupt_data", "write_failed"]


class TextStream(QtCore.QTextStream):
    def set_field_alignment(self, alignment: FieldAlignmentStr):
        """Set the field alignment.

        Args:
            alignment: field alignment

        Raises:
            InvalidParamError: invalid field alignment
        """
        if alignment not in FIELD_ALIGNMENT:
            raise InvalidParamError(alignment, FIELD_ALIGNMENT)
        self.setFieldAlignment(FIELD_ALIGNMENT[alignment])

    def get_field_alignment(self) -> FieldAlignmentStr:
        """Get current field alignment.

        Returns:
            current field alignment
        """
        return FIELD_ALIGNMENT.inverse[self.fieldAlignment()]

    def set_status(self, status: StatusStr):
        """Set the status.

        Args:
            status: status

        Raises:
            InvalidParamError: invalid status
        """
        if status not in STATUS:
            raise InvalidParamError(status, STATUS)
        self.setStatus(STATUS[status])

    def get_status(self) -> StatusStr:
        """Get current status.

        Returns:
            current status
        """
        return STATUS.inverse[self.status()]

    def set_real_number_notation(self, notation: RealNumberNotationStr):
        """Set the real number notation.

        Args:
            notation: real number notation

        Raises:
            InvalidParamError: invalid real number notation
        """
        if notation not in REAL_NUMBER_NOTATION:
            raise InvalidParamError(notation, REAL_NUMBER_NOTATION)
        self.setRealNumberNotation(REAL_NUMBER_NOTATION[notation])

    def get_real_number_notation(self) -> RealNumberNotationStr:
        """Get current real number notation.

        Returns:
            current real number notation
        """
        return REAL_NUMBER_NOTATION.inverse[self.realNumberNotation()]

    def set_codec(self, codec: Union[bytes, str]):
        if isinstance(codec, str):
            codec = codec.encode()
        self.setCodec(codec)

    def read_lines(self) -> Iterator[str]:
        while True:
            msg = self.readLine()
            if not msg:
                return
            yield msg


if __name__ == "__main__":
    stream = TextStream(QtCore.QByteArray(b"Test"))
    print(repr(stream))



from __future__ import annotations

from typing import Callable, Literal, Union

from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError, bidict


TYPE = bidict(
    linear=QtCore.QEasingCurve.Linear,
    in_quad=QtCore.QEasingCurve.InQuad,
    out_quad=QtCore.QEasingCurve.OutQuad,
    in_out_quad=QtCore.QEasingCurve.InOutQuad,
    out_in_quad=QtCore.QEasingCurve.OutInQuad,
    in_cubic=QtCore.QEasingCurve.InCubic,
    out_cubic=QtCore.QEasingCurve.OutCubic,
    in_out_cubic=QtCore.QEasingCurve.InOutCubic,
    out_in_cubic=QtCore.QEasingCurve.OutInCubic,
    in_quart=QtCore.QEasingCurve.InQuart,
    out_quart=QtCore.QEasingCurve.OutQuart,
    in_out_quart=QtCore.QEasingCurve.InOutQuart,
    out_in_quart=QtCore.QEasingCurve.OutInQuart,
    in_quint=QtCore.QEasingCurve.InQuint,
    out_quint=QtCore.QEasingCurve.OutQuint,
    in_out_quint=QtCore.QEasingCurve.InOutQuint,
    out_in_quint=QtCore.QEasingCurve.OutInQuint,
    in_sine=QtCore.QEasingCurve.InSine,
    out_sine=QtCore.QEasingCurve.OutSine,
    in_out_sine=QtCore.QEasingCurve.InOutSine,
    out_in_sine=QtCore.QEasingCurve.OutInSine,
    in_expo=QtCore.QEasingCurve.InExpo,
    out_expo=QtCore.QEasingCurve.OutExpo,
    in_out_expo=QtCore.QEasingCurve.InOutExpo,
    out_in_expo=QtCore.QEasingCurve.OutInExpo,
    in_circ=QtCore.QEasingCurve.InCirc,
    out_circ=QtCore.QEasingCurve.OutCirc,
    in_out_circ=QtCore.QEasingCurve.InOutCirc,
    out_in_circ=QtCore.QEasingCurve.OutInCirc,
    in_elastic=QtCore.QEasingCurve.InElastic,
    out_elastic=QtCore.QEasingCurve.OutElastic,
    in_out_elastic=QtCore.QEasingCurve.InOutElastic,
    out_in_elastic=QtCore.QEasingCurve.OutInElastic,
    in_back=QtCore.QEasingCurve.InBack,
    out_back=QtCore.QEasingCurve.OutBack,
    in_out_back=QtCore.QEasingCurve.InOutBack,
    out_in_back=QtCore.QEasingCurve.OutInBack,
    in_bounce=QtCore.QEasingCurve.InBounce,
    out_bounce=QtCore.QEasingCurve.OutBounce,
    in_out_bounce=QtCore.QEasingCurve.InOutBounce,
    out_in_bounce=QtCore.QEasingCurve.OutInBounce,
    bezier_spline=QtCore.QEasingCurve.BezierSpline,
    tcb_spline=QtCore.QEasingCurve.TCBSpline,
    custom=QtCore.QEasingCurve.Custom,
)

TypeStr = Literal[
    "linear",
    "in_quad",
    "out_quad",
    "in_out_quad",
    "out_in_quad",
    "in_cubic",
    "out_cubic",
    "in_out_cubic",
    "out_in_cubic",
    "in_quart",
    "out_quart",
    "in_out_quart",
    "out_in_quart",
    "in_quint",
    "out_quint",
    "in_out_quint",
    "out_in_quint",
    "in_sine",
    "out_sine",
    "in_out_sine",
    "out_in_sine",
    "in_expo",
    "out_expo",
    "in_out_expo",
    "out_in_expo",
    "in_circ",
    "out_circ",
    "in_out_circ",
    "out_in_circ",
    "in_elastic",
    "out_elastic",
    "in_out_elastic",
    "out_in_elastic",
    "in_back",
    "out_back",
    "in_out_back",
    "out_in_back",
    "in_bounce",
    "out_bounce",
    "in_out_bounce",
    "out_in_bounce",
    "bezier_spline",
    "tcb_spline",
    "custom",
]
CurveMethod = Callable[[float], float]


class EasingCurve(QtCore.QEasingCurve):
    def __init__(
        self, other_or_type: Union[TypeStr, int, QtCore.QEasingCurve] = "linear"
    ):
        if isinstance(other_or_type, str) and other_or_type in TYPE:
            typ = TYPE[other_or_type]
        else:
            typ = other_or_type
        super().__init__(typ)

    def __getitem__(self, value: float) -> float:
        return self.valueForProgress(value)

    def __repr__(self):
        return f"{type(self).__name__}({self.get_type()!r})"

    def set_custom_type(self, method: CurveMethod):
        self.setCustomType(method)

    def get_custom_type(self) -> CurveMethod:
        return self.customType()  # type: ignore

    def set_type(self, typ: TypeStr):
        """Set easing curve type.

        Args:
            typ: easing curve type

        Raises:
            InvalidParamError: easing curve type does not exist
        """
        if typ not in TYPE:
            raise InvalidParamError(typ, TYPE)
        self.setType(TYPE[typ])

    def get_type(self) -> TypeStr:
        """Get the current easing curve type.

        Returns:
            easing curve type
        """
        return TYPE.inverse[self.type()]


if __name__ == "__main__":
    c = EasingCurve()
    print(repr(c))



from __future__ import annotations

from typing import Iterator, Literal, Tuple, Union

from prettyqt import core
from prettyqt.qt import QtCore


class LineF(QtCore.QLineF):
    def __repr__(self):
        return f"{type(self).__name__}({repr(self.get_p1())}, {repr(self.get_p1())})"

    def __reduce__(self):
        return type(self), (self.get_p1(), self.get_p1())

    def __reversed__(self):
        return LineF(self.get_p2(), self.get_p1())

    def __abs__(self) -> float:
        return self.length()

    def __iter__(self) -> Iterator[core.PointF]:
        yield self.get_p1()
        yield self.get_p2()

    def __getitem__(self, index: Literal[0, 1]) -> core.PointF:
        if index == 0:
            return self.get_p1()
        elif index == 1:
            return self.get_p2()
        else:
            raise KeyError(index)

    def __setitem__(
        self, index: Literal[0, 1], value: Union[QtCore.QPointF, Tuple[float, float]]
    ):
        if index == 0:
            self.set_p1(value)
        elif index == 1:
            self.set_p2(value)
        else:
            raise KeyError(index)

    def set_p1(self, point: Union[QtCore.QPointF, Tuple[float, float]]):
        self.setP1(core.PointF(*point) if isinstance(point, tuple) else point)

    def get_p1(self) -> core.PointF:
        return core.PointF(self.p1())

    def set_p2(self, point: Union[QtCore.QPointF, Tuple[float, float]]):
        self.setP2(core.PointF(*point) if isinstance(point, tuple) else point)

    def get_p2(self) -> core.PointF:
        return core.PointF(self.p2())

    def get_center(self) -> core.PointF:
        return core.PointF(self.center())

    def get_normal_vector(self) -> LineF:
        return LineF(self.normalVector())

    def get_unit_vector(self) -> LineF:
        return LineF(self.unitVector())

    def to_line(self) -> core.Line:
        return core.Line(self.toLine())


if __name__ == "__main__":
    line = LineF(core.Point(0, 0), core.Point(2, 2))
    print(repr(line))
    for p in line:
        print(p)



from __future__ import annotations

from typing import Literal, Union

from prettyqt import core, gui
from prettyqt.qt.QtCharts import QtCharts
from prettyqt.utils import bidict


VALUE_POSITION = bidict(
    lower_extreme=QtCharts.QBoxSet.LowerExtreme,
    lower_quartile=QtCharts.QBoxSet.LowerQuartile,
    median=QtCharts.QBoxSet.Median,
    upper_quartile=QtCharts.QBoxSet.UpperQuartile,
    upper_extreme=QtCharts.QBoxSet.UpperExtreme,
)

ValuePositionStr = Literal[
    "lower_extreme", "lower_quartile", "median", "upper_quartile", "upper_extreme"
]

QtCharts.QBoxSet.__bases__ = (core.Object,)


class BoxSet(QtCharts.QBoxSet):
    def __repr__(self):
        return (
            f"{type(self).__name__}({self['lower_extreme']}, {self['lower_quartile']},"
            f" {self['median']}, {self['upper_quartile']}, {self['upper_extreme']}, "
            f"{self.label()!r})"
        )

    def __getitem__(self, index: Union[int, ValuePositionStr]) -> float:
        if isinstance(index, str):
            index = VALUE_POSITION[index]
        elif not (0 <= index <= 4):
            raise KeyError(index)
        return self.at(index)

    def __setitem__(self, index: Union[int, ValuePositionStr], value: int):
        if isinstance(index, str):
            index = VALUE_POSITION[index]
        elif not (0 <= index <= 4):
            raise KeyError(index)
        self.setValue(index, value)

    def get_pen(self) -> gui.Pen:
        return gui.Pen(self.pen())

    def get_brush(self) -> gui.Brush:
        return gui.Brush(self.brush())


if __name__ == "__main__":
    boxset = BoxSet()
    print(repr(boxset))



from __future__ import annotations

import os
import pathlib
from typing import List, Union

from prettyqt import core, widgets
from prettyqt.qt import QtWidgets


QtWidgets.QTextBrowser.__bases__ = (widgets.TextEdit,)


class TextBrowser(QtWidgets.QTextBrowser):

    value_changed = core.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setOpenExternalLinks(True)

    def serialize_fields(self):
        return dict(
            open_external_links=self.openExternalLinks(),
            open_links=self.openLinks(),
            search_paths=self.get_search_paths(),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setOpenExternalLinks(state["open_external_links"])
        self.setOpenLinks(state["open_links"])
        self.set_search_paths(state["search_paths"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    # def dragEnterEvent(self, event):
    #     u = event.mimeData().urls()
    #     for url in u:
    #         file_path = os.path.abspath(url.toLocalFile())

    #         ext = file_path.split(".")[-1]
    #         if ext in ["txt", "md", "markdown"]:
    #             event.accept()
    #         else:
    #             event.ignore()

    # def dropEvent(self, event):
    #     event.accept()
    #     self.show_markdown_file(self.filePath)

    def set_markdown_file(self, file_path: Union[str, os.PathLike]):
        file_path = pathlib.Path(file_path)
        with file_path.open() as f:
            file_content = f.read()
        self.set_markdown(file_content)

    def set_markdown(self, source: str):
        self.setMarkdown(source)

    def set_rst_file(self, file_path: Union[str, os.PathLike]):
        file_path = pathlib.Path(file_path)
        with file_path.open() as f:
            file_content = f.read()
        self.set_rst(file_content)

    def set_rst(self, source: str):
        import docutils.core

        html = docutils.core.publish_string(source, writer_name="html")
        self.setHtml(str(html))

    def get_search_paths(self) -> List[pathlib.Path]:
        return [pathlib.Path(p) for p in self.searchPaths()]

    def set_search_paths(self, paths: List[Union[str, os.PathLike]]):
        self.setSearchPaths([os.fspath(p) for p in paths])


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    reader = TextBrowser()
    reader.show()
    app.main_loop()



"""Module containing helper functions."""

from typing import Iterable
import logging

import bidict as bdct


logger = logging.getLogger(__name__)


class bidict(bdct.bidict):
    def __init__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            super().__init__(args[0])
        else:
            super().__init__(kwargs)


class InvalidParamError(ValueError):
    """Exception raised for invalid params in method calls.

    Args:
        value: param value which caused the error
        valid_options: allowed options
    """

    def __init__(self, value, valid_options: Iterable):
        self.value = value
        opts = " / ".join(repr(opt) for opt in valid_options)
        self.message = f"Invalid value: {value!r}. Allowed options are {opts}."
        super().__init__(self.message)



from __future__ import annotations

from prettyqt import constants, gui
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError


class PainterPathStroker(QtGui.QPainterPathStroker):
    def set_cap_style(self, style: constants.CapStyleStr):
        """Set cap style to use.

        Args:
            style: cap style to use

        Raises:
            InvalidParamError: cap style does not exist
        """
        if style not in constants.CAP_STYLE:
            raise InvalidParamError(style, constants.CAP_STYLE)
        self.setCapStyle(constants.CAP_STYLE[style])

    def get_cap_style(self) -> constants.CapStyleStr:
        """Return current cap style.

        Returns:
            cap style
        """
        return constants.CAP_STYLE.inverse[self.capStyle()]

    def set_join_style(self, style: constants.JoinStyleStr):
        """Set join style to use.

        Args:
            style: join style to use

        Raises:
            InvalidParamError: join style does not exist
        """
        if style not in constants.JOIN_STYLE:
            raise InvalidParamError(style, constants.JOIN_STYLE)
        self.setJoinStyle(constants.JOIN_STYLE[style])

    def get_join_style(self) -> constants.JoinStyleStr:
        """Return current join style.

        Returns:
            join style
        """
        return constants.JOIN_STYLE.inverse[self.joinStyle()]

    def create_stroke(self, path: QtGui.QPainterPath) -> gui.PainterPath:
        return gui.PainterPath(self.createStroke(path))


if __name__ == "__main__":
    p = PainterPathStroker(QtCore.QPoint(1, 1))
    print(list(p))



"""winextras module.

contains QtWinExtras-based classes
"""

try:
    from .winjumplistcategory import WinJumpListCategory
    from .winjumplistitem import WinJumpListItem
    from .winjumplist import WinJumpList
    from .winthumbnailtoolbutton import WinThumbnailToolButton

    __all__ = [
        "WinJumpListCategory",
        "WinJumpListItem",
        "WinJumpList",
        "WinThumbnailToolButton",
    ]
except (ImportError, ModuleNotFoundError):
    pass



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError, bidict


DELETION_POLICY = bidict(
    keep=QtCore.QAbstractAnimation.KeepWhenStopped,
    delete=QtCore.QAbstractAnimation.DeleteWhenStopped,
)

DeletionPolicyStr = Literal["keep", "delete"]

DIRECTION = bidict(
    forward=QtCore.QAbstractAnimation.Forward, backward=QtCore.QAbstractAnimation.Backward
)

DirectionStr = Literal["forward", "backward"]

STATE = bidict(
    stopped=QtCore.QAbstractAnimation.Stopped,
    paused=QtCore.QAbstractAnimation.Paused,
    running=QtCore.QAbstractAnimation.Running,
)

StateStr = Literal["stopped", "paused", "running"]

QtCore.QAbstractAnimation.__bases__ = (core.Object,)


class AbstractAnimation(QtCore.QAbstractAnimation):
    def __len__(self):
        return self.duration()

    def __and__(self, other: AbstractAnimation) -> core.SequentialAnimationGroup:
        group = core.SequentialAnimationGroup()
        group.addAnimation(self)
        group.addAnimation(other)
        return group

    def __or__(self, other: AbstractAnimation) -> core.ParallelAnimationGroup:
        group = core.ParallelAnimationGroup()
        group.addAnimation(self)
        group.addAnimation(other)
        return group

    def serialize_fields(self):
        return dict(
            duration=self.duration(),
            direction=self.get_direction(),
            loop_count=self.loopCount(),
            current_time=self.currentTime(),
        )

    def set_direction(self, direction: DirectionStr):
        """Set animation direction.

        Args:
            direction: animation direction

        Raises:
            InvalidParamError: animation direction does not exist
        """
        if direction not in DIRECTION:
            raise InvalidParamError(direction, DIRECTION)
        self.setDirection(DIRECTION[direction])

    def get_direction(self) -> DirectionStr:
        """Get the current animation direction.

        Returns:
            animation direction
        """
        return DIRECTION.inverse[self.direction()]

    def get_state(self) -> StateStr:
        """Get the current animation state.

        Returns:
            animation state
        """
        return STATE.inverse[self.state()]

    def start_animation(self, policy: DeletionPolicyStr = "keep"):
        """Start the animation.

        Args:
            policy: animation policy

        Raises:
            InvalidParamError: animation policy does not exist
        """
        if policy not in DELETION_POLICY:
            raise InvalidParamError(policy, DELETION_POLICY)
        self.start(DELETION_POLICY[policy])

    def restart_animation(self, policy: DeletionPolicyStr = "keep"):
        """Restart the animation.

        Args:
            policy: animation policy

        Raises:
            InvalidParamError: animation policy does not exist
        """
        self.stop()
        self.start_animation(policy)


if __name__ == "__main__":
    anim1 = core.PropertyAnimation()
    anim2 = core.PropertyAnimation()
    group = anim1 | anim2
    print(type(group))



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtWebEngineWidgets
from prettyqt.utils import InvalidParamError, bidict


INJECTION_POINT = bidict(
    document_creation=QtWebEngineWidgets.QWebEngineScript.DocumentCreation,
    document_ready=QtWebEngineWidgets.QWebEngineScript.DocumentReady,
    deferred=QtWebEngineWidgets.QWebEngineScript.Deferred,
)

InjectionPointStr = Literal["document_creation", "document_ready", "deferred"]

SCRIPT_WORLD_IDS = bidict(
    main_world=QtWebEngineWidgets.QWebEngineScript.MainWorld,
    application_world=QtWebEngineWidgets.QWebEngineScript.ApplicationWorld,
    user_world=QtWebEngineWidgets.QWebEngineScript.UserWorld,
)

ScriptWorldIdStr = Literal["main_world", "application_world", "user_world"]


class WebEngineScript(QtWebEngineWidgets.QWebEngineScript):
    def set_injection_point(self, point: InjectionPointStr):
        """Set injection point.

        Args:
            point: injection point to use

        Raises:
            InvalidParamError: injection point does not exist
        """
        if point not in INJECTION_POINT:
            raise InvalidParamError(point, INJECTION_POINT)
        self.setInjectionPoint(INJECTION_POINT[point])

    def get_injection_point(self) -> InjectionPointStr:
        """Return injection point.

        Returns:
            injection point
        """
        return INJECTION_POINT.inverse[self.injectionPoint()]


if __name__ == "__main__":
    item = WebEngineScript()



from __future__ import annotations

from typing import List, Optional, Tuple, Union

from prettyqt import core, location, positioning
from prettyqt.qt import QtLocation, QtPositioning


QtLocation.QPlaceManager.__bases__ = (core.Object,)


class PlaceManager(core.Object):
    on_finished = core.Signal(location.PlaceSearchReply)

    def __init__(self, item: QtLocation.QPlaceManager):
        super().__init__()
        self.item = item
        self.finished.connect(self._on_finished)

    def __getattr__(self, val):
        return getattr(self.item, val)

    def _on_finished(self, reply: QtLocation.QPlaceSearchReply):
        reply = location.PlaceSearchReply.clone_from(reply)
        self.on_finished.emit(reply)

    def get_category(self, cat_id: str) -> location.PlaceCategory:
        return location.PlaceCategory(self.item.category(cat_id))

    def get_child_categories(self, cat_id: str) -> List[location.PlaceCategory]:
        return [location.PlaceCategory(i) for i in self.item.childCategories(cat_id)]

    def get_locales(self) -> List[core.Locale]:
        return [core.Locale(i) for i in self.locales()]

    def search_place(
        self,
        search_term: str,
        coord: Union[Tuple[float, float], QtPositioning.QGeoCoordinate],
        radius: Optional[float] = None,
        limit: Optional[int] = None,
        relevance: Optional[location.placesearchrequest.RelevanceHintStr] = None,
        categories: Optional[List[str]] = None,
    ):
        request = location.PlaceSearchRequest()
        request.setSearchTerm(search_term)
        if radius is None:
            radius = -1
        if isinstance(coord, tuple):
            coord = positioning.GeoCoordinate(*coord)
        circle = positioning.GeoCircle(coord, radius)
        request.setSearchArea(circle)
        if limit is not None:
            request.setLimit(limit)
        if relevance is not None:
            request.set_relevance_hint(relevance)
        if categories is not None:
            self.setCategories(categories)
        return self.search(request)


if __name__ == "__main__":
    app = core.app()
    provider = location.GeoServiceProvider("osm")
    manager = provider.get_place_manager()
    reply = manager.search_place("Shop", coord=(51, 7))
    manager.on_finished.connect(lambda x: print(x.get_results()))
    app.main_loop()



from __future__ import annotations

from typing import Callable

from prettyqt import constants, core
from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError


QtCore.QTimer.__bases__ = (core.Object,)


class Timer(QtCore.QTimer):
    @classmethod
    def single_shot(cls, callback: Callable) -> Timer:
        timer = cls()
        timer.timeout.connect(callback)
        timer.setSingleShot(True)
        return timer

    def set_type(self, typ: constants.TimerTypeStr):
        """Set the timer type.

        Args:
            typ: timer type

        Raises:
            InvalidParamError: timer type does not exist
        """
        if typ not in constants.TIMER_TYPE:
            raise InvalidParamError(typ, constants.TIMER_TYPE)
        self.setTimerType(constants.TIMER_TYPE[typ])

    def get_type(self) -> constants.TimerTypeStr:
        """Return current timer type.

        Returns:
            timer type
        """
        return constants.TIMER_TYPE.inverse[self.timerType()]

    def restart(self):
        self.stop()
        self.start()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtCore, QtWidgets


QtWidgets.QTreeWidget.__bases__ = (widgets.TreeView,)


class TreeWidget(QtWidgets.QTreeWidget):
    def __contains__(self, other: QtWidgets.QTreeWidgetItem):
        return self.indexOfTopLevelItem(other) >= 0

    def sort(self, column: int = 0, reverse: bool = False):
        order = QtCore.Qt.DescendingOrder if reverse else QtCore.Qt.AscendingOrder
        self.sortItems(column, order)


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    widget = TreeWidget()
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QWidgetItem.__bases__ = (widgets.LayoutItem,)


class WidgetItem(QtWidgets.QWidgetItem):
    pass


if __name__ == "__main__":
    item = WidgetItem(QtWidgets.QWidget())



from __future__ import annotations

from typing import Literal, Tuple, Union

from prettyqt import core, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict


VIEW_MODE = bidict(list=QtWidgets.QListView.ListMode, icon=QtWidgets.QListView.IconMode)

ViewModeStr = Literal["list", "icon"]

MOVEMENT = bidict(
    static=QtWidgets.QListView.Static,
    free=QtWidgets.QListView.Free,
    snap=QtWidgets.QListView.Snap,
)

MovementStr = Literal["static", "free", "snap"]

FLOW = bidict(
    left_to_right=QtWidgets.QListView.LeftToRight,
    top_to_bottom=QtWidgets.QListView.TopToBottom,
)

FlowStr = Literal["left_to_right", "top_to_bottom"]

LAYOUT_MODE = bidict(
    single_pass=QtWidgets.QListView.SinglePass,
    batched=QtWidgets.QListView.Batched,
)

LayoutModeStr = Literal["single_pass", "batched"]

RESIZE_MODE = bidict(
    fixed=QtWidgets.QListView.Fixed,
    adjust=QtWidgets.QListView.Adjust,
)

ResizeModeStr = Literal["fixed", "adjust"]


QtWidgets.QListView.__bases__ = (widgets.AbstractItemView,)


class ListView(QtWidgets.QListView):
    def serialize_fields(self):
        return dict(
            view_mode=self.get_view_mode(),
            resize_mode=self.get_resize_mode(),
            layout_mode=self.get_layout_mode(),
            movement=self.get_movement(),
            batch_size=self.batchSize(),
            flow=self.get_flow(),
            grid_size=self.get_grid_size(),
            is_wrapping=self.isWrapping(),
            # item_alignment=self.itemAlignment(),
            model_column=self.modelColumn(),
            selection_rect_visible=self.isSelectionRectVisible(),
            spacing=self.spacing(),
            uniform_item_sizes=self.uniformItemSizes(),
            word_wrap=self.wordWrap(),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self.set_view_mode(state["view_mode"])
        self.set_resize_mode(state["resize_mode"])
        self.set_layout_mode(state["layout_mode"])
        self.set_movement(state["movement"])
        self.setBatchSize(state["batch_size"])
        self.set_flow(state["flow"])
        self.set_grid_size(state["grid_size"])
        self.setWrapping(state["is_wrapping"])
        # self.setItemAlignment(state["flow"])
        self.setModelColumn(state["model_column"])
        self.setSelectionRectVisible(state["selection_rect_visible"])
        self.setSpacing(state["spacing"])
        self.setUniformItemSizes(state["uniform_item_sizes"])
        self.setWordWrap(state["word_wrap"])

    def set_view_mode(self, mode: ViewModeStr):
        """Set view mode.

        Args:
            mode: view mode to use

        Raises:
            InvalidParamError: invalid view mode
        """
        if mode not in VIEW_MODE:
            raise InvalidParamError(mode, VIEW_MODE)
        self.setViewMode(VIEW_MODE[mode])

    def get_view_mode(self) -> ViewModeStr:
        """Return view mode.

        Returns:
            view mode
        """
        return VIEW_MODE.inverse[self.viewMode()]

    def set_resize_mode(self, mode: ResizeModeStr):
        """Set resize mode.

        Args:
            mode: resize mode to use

        Raises:
            InvalidParamError: invalid resize mode
        """
        if mode not in RESIZE_MODE:
            raise InvalidParamError(mode, RESIZE_MODE)
        self.setResizeMode(RESIZE_MODE[mode])

    def get_resize_mode(self) -> ResizeModeStr:
        """Return resize mode.

        Returns:
            resize mode
        """
        return RESIZE_MODE.inverse[self.resizeMode()]

    def set_layout_mode(self, mode: LayoutModeStr):
        """Set layout mode.

        Args:
            mode: layout mode to use

        Raises:
            InvalidParamError: invalid layout mode
        """
        if mode not in LAYOUT_MODE:
            raise InvalidParamError(mode, LAYOUT_MODE)
        self.setLayoutMode(LAYOUT_MODE[mode])

    def get_layout_mode(self) -> LayoutModeStr:
        """Return layout mode.

        Returns:
            layout mode
        """
        return LAYOUT_MODE.inverse[self.layoutMode()]

    def set_movement(self, mode: MovementStr):
        """Set movement mode.

        Args:
            mode: movement mode to use

        Raises:
            InvalidParamError: invalid movement mode
        """
        if mode not in MOVEMENT:
            raise InvalidParamError(mode, MOVEMENT)
        self.setMovement(MOVEMENT[mode])

    def get_movement(self) -> MovementStr:
        """Return movement mode.

        Returns:
            movement mode
        """
        return MOVEMENT.inverse[self.movement()]

    def set_flow(self, mode: FlowStr):
        """Set flow mode.

        Args:
            mode: flow mode to use

        Raises:
            InvalidParamError: invalid flow mode
        """
        if mode not in FLOW:
            raise InvalidParamError(mode, FLOW)
        self.setFlow(FLOW[mode])

    def get_flow(self) -> FlowStr:
        """Return flow mode.

        Returns:
            flow mode
        """
        return FLOW.inverse[self.flow()]

    def set_grid_size(self, size: Union[QtCore.QSize, Tuple[int, int]]):
        if isinstance(size, tuple):
            size = QtCore.QSize(*size)
        self.setGridSize(size)

    def get_grid_size(self) -> core.Size:
        return core.Size(self.gridSize())


if __name__ == "__main__":
    app = widgets.app()
    dlg = ListView()
    dlg.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStyleOptionFocusRect.__bases__ = (widgets.StyleOption,)


class StyleOptionFocusRect(QtWidgets.QStyleOptionFocusRect):
    pass



# Functions for the Qt meta-object system (like signal, slot, etc)
# credits to Empyrical
# https://gist.github.com/empyrical/a103c46f454a433e8798#file-decorator-py

from __future__ import annotations

import inspect
import logging
from typing import Callable, Optional, Tuple, get_type_hints

from prettyqt import core


logger = logging.getLogger(__name__)


class _OverloadedSlotPlaceholder:
    def __init__(self, name, original_name):
        self.name = name
        self.original_name = original_name

    def __repr__(self):
        return f"<placeholder slot {self.name} for function {self.original_name}>"


def _build_arguments(
    func: Callable, original_func: Optional[Callable] = None
) -> Tuple[list, dict]:
    argspec = inspect.getfullargspec(func)
    args = argspec.args
    annotations = get_type_hints(func)

    slot_args = []
    slot_kwargs = {}

    if "return" in annotations and annotations["return"] is not None:
        slot_kwargs["result"] = annotations["return"]

    if original_func and func.__name__ != original_func.__name__:
        slot_kwargs["name"] = func.__name__

    gap = False

    for index, arg in enumerate(args):
        # Let the first arg pass by unannotated, as it may be the 'self'
        # argument of a function. Static methods can't be used as slots anyways
        if index == 0 and arg not in annotations:
            continue

        if arg in annotations and not gap:
            slot_args.append(annotations[arg])
        elif arg not in annotations:
            gap = True
        elif arg in annotations and gap:
            raise TypeError(
                "Type annotations must be in a continuous row - an argument "
                f'before "{arg}" is missing'
            )

    return slot_args, slot_kwargs


def AutoSlot(func: Callable):
    """A special slot decorator using function annotations to provide the argument types.

    It's just syntactic sugar for the slot decorators that already come with the
    Python Qt bindings.

    This has **no effect whatsoever** on the Python side of things - it does not
    do type checking, it only registers a function and the types its arguments
    take with Qt's Meta Object system.

    Use like so:

        @slot
        def my_function(self, x: int, y: int) -> int:
            return x ** y

    You can also overload a function by doing this:

        @slot
        def overloaded_func(self, output=None: str):
            print(output)

        @overloaded_func.overload
        def overloaded_func(self, output: int): ...

        # And an overload for calling it with no arguments, so it will fall back
        # on the default argument.
        @overloaded_func.overload
        def overloaded_func(self): ...

    Note: you cannot set different default arguments in the overloads, you may
    only set a function's default arguments in the main declaration.

    The function body on overloads is completely ignored, so you can just use
    ellipses (`...`) or `pass` to fill in the body.
    """
    args, kwargs = _build_arguments(func)
    logger.debug(f"Auto-slot for {func.__name__}: args: {args}, kwargs: {kwargs}")

    core.Slot(*args, **kwargs)(func)

    def overload(new_func):
        args, kwargs = _build_arguments(new_func, func)

        core.Slot(*args, **kwargs)(func)

        if new_func.__name__ != func.__name__:
            return _OverloadedSlotPlaceholder(new_func.__name__, func.__name__)
        else:
            return func

    func.overload = overload  # type: ignore

    return func



from __future__ import annotations

from typing import Optional

from prettyqt import constants, core, widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError


QtWidgets.QCheckBox.__bases__ = (widgets.AbstractButton,)


class CheckBox(QtWidgets.QCheckBox):

    value_changed = core.Signal(int)

    def __init__(
        self,
        label: str = "",
        parent: Optional[QtWidgets.QWidget] = None,
        checked: bool = False,
    ):
        super().__init__(label, parent)
        self.stateChanged.connect(self.value_changed)
        self.setChecked(checked)

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setTristate(state.get("is_tristate", False))
        self.set_checkstate(state["checkstate"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def serialize_fields(self):
        return dict(
            checkstate=self.get_checkstate(),
            is_tristate=self.isTristate(),
        )

    def set_checkstate(self, state: constants.StateStr):
        """Set checkstate of the checkbox.

        Args:
            state: checkstate to use

        Raises:
            InvalidParamError: invalid checkstate
        """
        if state not in constants.STATE:
            raise InvalidParamError(state, constants.STATE)
        self.setCheckState(constants.STATE[state])

    def get_checkstate(self) -> constants.StateStr:
        """Return checkstate.

        Returns:
            checkstate
        """
        return constants.STATE.inverse[self.checkState()]


if __name__ == "__main__":
    app = widgets.app()
    widget = CheckBox("test")
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QAbstractTableModel.__bases__ = (core.AbstractItemModel,)


class AbstractTableModel(QtCore.QAbstractTableModel):
    pass



from __future__ import annotations

from typing import Union

from prettyqt import core
from prettyqt.qt import QtCore
from prettyqt.utils import bidict


SYSTEM = bidict(
    gregorian=QtCore.QCalendar.System.Gregorian,
    julian=QtCore.QCalendar.System.Julian,
    milankovic=QtCore.QCalendar.System.Milankovic,
    jalali=QtCore.QCalendar.System.Jalali,
    islamic_civil=QtCore.QCalendar.System.IslamicCivil,
)


class Calendar(QtCore.QCalendar):
    def __init__(self, system: Union[str, QtCore.QCalendar.System] = "gregorian"):
        if isinstance(system, QtCore.QCalendar.System):
            typ = system
        else:
            typ = SYSTEM[system]
        super().__init__(typ)

    def __repr__(self):
        return f"{type(self).__name__}({self.name()!r})"

    def __reduce__(self):
        return type(self), (self.name(),)

    # def __bool__(self):
    #     return self.isValid()

    def get_date_from_parts(self, year: int, month: int, day: int) -> core.Date:
        return core.Date(self.dateFromParts(year, month, day))


if __name__ == "__main__":
    cal = Calendar("gregorian")
    print(repr(cal))



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtHelp


QtHelp.QHelpSearchQueryWidget.__bases__ = (widgets.Widget,)


class HelpSearchQueryWidget(QtHelp.QHelpSearchQueryWidget):
    pass



# type: ignore
# not available in PySide2

from __future__ import annotations

from typing import List, Literal

from prettyqt import core
from prettyqt.qt import QtCore
from prettyqt.utils import bidict


LOAD_HINTS = bidict(
    resolve_all_symbols=QtCore.QLibrary.ResolveAllSymbolsHint,
    export_external_symbols=QtCore.QLibrary.ExportExternalSymbolsHint,
    load_archive_member=QtCore.QLibrary.LoadArchiveMemberHint,
    prevent_unload=QtCore.QLibrary.PreventUnloadHint,
    deep_bind=QtCore.QLibrary.DeepBindHint,
)

LoadHintStr = Literal[
    "resolve_all_symbols",
    "export_external_symbols",
    "load_archive_member",
    "prevent_unload",
    "deep_bind",
]

QtCore.QLibrary.__bases__ = (core.Object,)


class Library(QtCore.QLibrary):
    def __bool__(self):
        return self.isLoaded()

    def __repr__(self):
        return f"{type(self).__name__}({self.fileName()!r})"

    def get_load_hints(self) -> List[LoadHintStr]:
        return [k for k, v in LOAD_HINTS.items() if v & self.loadHints()]

    def set_load_hints(self, **kwargs):
        flag = QtCore.QLibrary.LoadHint(0)
        for k, v in kwargs.items():
            if v is True:
                flag |= LOAD_HINTS[k]
        self.setLoadHints(flag)



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QPauseAnimation.__bases__ = (core.AbstractAnimation,)


class PauseAnimation(QtCore.QPauseAnimation):
    def __repr__(self):
        return f"{type(self).__name__}({self.duration()})"



from __future__ import annotations

from typing import Optional

from prettyqt import core, gui, widgets
from prettyqt.qt import QtCore, QtWidgets


class CodeEditor(widgets.PlainTextEdit):
    def __init__(
        self, language: str = "python", parent: Optional[QtWidgets.QWidget] = None
    ):
        super().__init__(parent=parent)
        self.line_area = LineNumberArea(self)
        self.blockCountChanged.connect(self.update_line_area_width)
        self.updateRequest.connect(self.update_line_area)
        self.cursorPositionChanged.connect(self.highlight_current_line)
        self.set_font("Consolas")
        self.update_line_area_width(0)
        self.highlight_current_line()
        self.set_syntaxhighlighter(language)

    def resizeEvent(self, event):
        super().resizeEvent(event)

        cr = self.contentsRect()
        rect = core.Rect(cr.left(), cr.top(), self.line_area_width(), cr.height())
        self.line_area.setGeometry(rect)

    def text(self) -> str:
        return self.toPlainText()

    def line_area_width(self) -> int:
        digits = len(str(self.blockCount()))
        return 3 + self.fontMetrics().boundingRect("9").width() * digits

    def update_line_area_width(self, _):
        self.setViewportMargins(self.line_area_width(), 0, 0, 0)

    def update_line_area(self, rect: QtCore.QRect, dy: int):
        if dy:
            self.line_area.scroll(0, dy)
        else:
            self.line_area.update(0, rect.y(), self.line_area.width(), rect.height())

        if rect.contains(self.viewport().rect()):
            self.update_line_area_width(0)

    def line_area_paintevent(self, event):
        with gui.Painter(self.line_area) as painter:
            painter.fill_rect(event.rect(), "lightgray")

            block = self.firstVisibleBlock()
            block_number = block.blockNumber()
            top = self.blockBoundingGeometry(block).translated(self.contentOffset()).top()
            bottom = top + self.blockBoundingRect(block).height()
            width = self.line_area.width()
            height = self.fontMetrics().height()
            painter.set_color("black")
            while block.isValid() and (top <= event.rect().bottom()):
                if block.isVisible() and (bottom >= event.rect().top()):
                    number = str(block_number + 1)
                    painter.drawText(
                        0, int(top), width, height, QtCore.Qt.AlignRight, number
                    )
                block = block.next()
                top = bottom
                bottom = top + self.blockBoundingRect(block).height()
                block_number += 1


class LineNumberArea(widgets.Widget):
    def __init__(self, editor: CodeEditor):
        super().__init__(editor)
        self.editor = editor

    def sizeHint(self) -> core.Size:
        return core.Size(self.editor.line_area_width(), 0)

    def paintEvent(self, event):
        self.editor.line_area_paintevent(event)


if __name__ == "__main__":
    app = widgets.app()
    dlg = CodeEditor()
    dlg.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsSimpleTextItem.__bases__ = (widgets.AbstractGraphicsShapeItem,)


class GraphicsSimpleTextItem(QtWidgets.QGraphicsSimpleTextItem):
    def __repr__(self):
        return f"{type(self).__name__}({self.text()!r})"



from __future__ import annotations

from prettyqt import constants, core, multimedia, widgets
from prettyqt.qt import QtMultimediaWidgets
from prettyqt.utils import InvalidParamError


QtMultimediaWidgets.QGraphicsVideoItem.__bases__ = (
    widgets.GraphicsObject,
    multimedia.MediaBindableInterface,
)


class GraphicsVideoItem(QtMultimediaWidgets.QGraphicsVideoItem):
    def get_offset(self) -> core.PointF:
        return core.PointF(self.offset())

    def get_native_size(self) -> core.SizeF:
        return core.SizeF(self.nativeSize())

    def get_size(self) -> core.SizeF:
        return core.SizeF(self.size())

    def set_aspect_ratio_mode(self, mode: constants.AspectRatioModeStr):
        """Set the aspect ratio mode.

        Args:
            mode: aspect ratio mode

        Raises:
            InvalidParamError: aspect ratio mode does not exist
        """
        if mode not in constants.ASPECT_RATIO_MODE:
            raise InvalidParamError(mode, constants.ASPECT_RATIO_MODE)
        self.setAspectRatioMode(constants.ASPECT_RATIO_MODE[mode])

    def get_aspect_ratio_mode(self) -> constants.AspectRatioModeStr:
        """Return current aspect ratio mode.

        Returns:
            aspect ratio mode
        """
        return constants.ASPECT_RATIO_MODE.inverse[self.aspectRatioMode()]



from __future__ import annotations

from prettyqt.qt import QtWidgets


class ButtonGroup(QtWidgets.QButtonGroup):
    def serialize_fields(self):
        return dict(exclusive=self.exclusive())

    def __getitem__(self, index: int) -> QtWidgets.QAbstractButton:
        return self.button(index)



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStyledItemDelegate.__bases__ = (widgets.AbstractItemDelegate,)


class StyledItemDelegate(QtWidgets.QStyledItemDelegate):
    pass



from __future__ import annotations

from typing import Iterator

from prettyqt.qt import QtCore


class Margins(QtCore.QMargins):
    def __repr__(self):
        return (
            f"{type(self).__name__}({self.left()}, {self.top()}, "
            f"{self.right()}, {self.bottom()})"
        )

    def __reduce__(self):
        return type(self), (self.left(), self.top(), self.right(), self.bottom())

    def __bool__(self):
        return not self.isNull()

    def __iter__(self) -> Iterator[int]:
        yield self.left()
        yield self.top()
        yield self.right()
        yield self.bottom()


if __name__ == "__main__":
    margins = Margins(0, 0, 0, 0)
    print(bool(margins))



from __future__ import annotations

import logging

from prettyqt import core, widgets


logger = logging.getLogger(__name__)


class CheckBoxDelegate(widgets.ItemDelegate):
    """Delegate that places a CheckBox in every cell."""

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.selected = []

    def createEditor(self, parent, option, index):
        """Override.

        instanciate the editor widget and initialize it
        also connect currentIndexChanged signal.
        """
        cb = widgets.CheckBox(parent)
        cb.currentIndexChanged.connect(self.currentIndexChanged)
        return cb

    def setEditorData(self, cb, index):
        """Override.

        set correct initial value for editor widget
        """
        current_selection = index.data()
        with cb.block_signals():
            cb.setCurrentText(current_selection)

    @core.Slot()
    def currentIndexChanged(self):
        self.commitData.emit(self.sender())



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtHelp


QtHelp.QHelpIndexWidget.__bases__ = (widgets.ListView,)


class HelpIndexWidget(QtHelp.QHelpIndexWidget):
    pass



from __future__ import annotations

from typing import Iterator, List

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsLayout.__bases__ = (widgets.GraphicsLayoutItem,)


class GraphicsLayout(QtWidgets.QGraphicsLayout):
    def __getitem__(self, index: int) -> QtWidgets.QGraphicsItem:
        layoutitem = self.itemAt(index)
        return layoutitem.graphicsItem()

    def __setitem__(self, index: int, value: QtWidgets.QGraphicsItem):
        layoutitem = self.itemAt(index)
        layoutitem.setGraphicsItem(value)

    def __delitem__(self, index: int):
        self.removeAt(index)

    def __iter__(self) -> Iterator[QtWidgets.QGraphicsItem]:
        return iter(self[i] for i in range(self.count()))

    def __contains__(self, item):
        return item in self.get_children()

    def __len__(self):
        # for PySide2
        return self.count()

    def get_children(self) -> List[QtWidgets.QGraphicsItem]:
        return list(self)

    def set_margin(self, margin: int):
        self.setContentsMargins(margin, margin, margin, margin)


if __name__ == "__main__":
    layout = GraphicsLayout()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStyleOptionProgressBar.__bases__ = (widgets.StyleOption,)


class StyleOptionProgressBar(QtWidgets.QStyleOptionProgressBar):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QSequentialAnimationGroup.__bases__ = (core.AnimationGroup,)


class SequentialAnimationGroup(QtCore.QSequentialAnimationGroup):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QItemSelectionModel.__bases__ = (core.Object,)


class ItemSelectionModel(QtCore.QItemSelectionModel):
    pass



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QScrollArea.__bases__ = (widgets.AbstractScrollArea,)


class ScrollArea(QtWidgets.QScrollArea):
    def serialize_fields(self):
        return dict(widget=self.widget(), resizable=self.widgetResizable())

    def __setstate__(self, state):
        super().__setstate__(state)
        self.set_widget(state["widget"])
        self.setWidgetResizable(state["resizable"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def set_widget(self, widget: QtWidgets.QWidget):
        self.setWidget(widget)



from __future__ import annotations

from typing import Literal, Union

from prettyqt import gui
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import bidict


ICON_TYPE = bidict(
    computer=QtWidgets.QFileIconProvider.Computer,
    desktop=QtWidgets.QFileIconProvider.Desktop,
    trashcan=QtWidgets.QFileIconProvider.Trashcan,
    network=QtWidgets.QFileIconProvider.Network,
    drive=QtWidgets.QFileIconProvider.Drive,
    folder=QtWidgets.QFileIconProvider.Folder,
    file=QtWidgets.QFileIconProvider.File,
)

IconTypeStr = Literal[
    "computer", "desktop", "trashcan", "network", "drive", "folder", "file"
]

OPTIONS = bidict(
    dont_use_custom_dir_icons=QtWidgets.QFileIconProvider.DontUseCustomDirectoryIcons,
)

OptionStr = Literal["dont_use_custom_dir_icons"]


class FileIconProvider(QtWidgets.QFileIconProvider):
    def get_icon(self, typ: Union[IconTypeStr, QtCore.QFileInfo]) -> gui.Icon:
        if not isinstance(typ, QtCore.QFileInfo):
            param = ICON_TYPE[typ]
        else:
            param = typ
        return gui.Icon(self.icon(param))



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtMultimedia


QtMultimedia.QMediaControl.__bases__ = (core.Object,)


class MediaControl(QtMultimedia.QMediaControl):
    pass



from __future__ import annotations

from typing import List, Literal

from prettyqt import core, positioning
from prettyqt.qt import QtLocation
from prettyqt.utils import InvalidParamError, bidict, helpers, mappers


FEATURE_TYPES = bidict(
    none=QtLocation.QGeoRouteRequest.NoFeature,
    toll=QtLocation.QGeoRouteRequest.TollFeature,
    highway=QtLocation.QGeoRouteRequest.HighwayFeature,
    public_transit=QtLocation.QGeoRouteRequest.PublicTransitFeature,
    ferry=QtLocation.QGeoRouteRequest.FerryFeature,
    tunnel=QtLocation.QGeoRouteRequest.TunnelFeature,
    dirt_road=QtLocation.QGeoRouteRequest.DirtRoadFeature,
    parks=QtLocation.QGeoRouteRequest.ParksFeature,
    motor_pool_lane=QtLocation.QGeoRouteRequest.MotorPoolLaneFeature,
    traffic=QtLocation.QGeoRouteRequest.TrafficFeature,
)

FeatureTypeStr = Literal[
    "none",
    "toll",
    "highway",
    "public_transit",
    "ferry",
    "tunnel",
    "dirt_road",
    "parks",
    "motor_pool_lane",
    "traffic",
]

FEATURE_WEIGHTS = bidict(
    neutral=QtLocation.QGeoRouteRequest.NeutralFeatureWeight,
    prefer=QtLocation.QGeoRouteRequest.PreferFeatureWeight,
    require=QtLocation.QGeoRouteRequest.RequireFeatureWeight,
    avoid=QtLocation.QGeoRouteRequest.AvoidFeatureWeight,
    disallow=QtLocation.QGeoRouteRequest.DisallowFeatureWeight,
)

FeatureWeightStr = Literal["neutral", "prefer", "require", "avoid", "disallow"]

MANEUVER_DETAIL = bidict(
    none=QtLocation.QGeoRouteRequest.NoManeuvers,
    basic=QtLocation.QGeoRouteRequest.BasicManeuvers,
)

ManeuverDetailStr = Literal["none", "basic"]

ROUTE_OPTIMIZATION = mappers.FlagMap(
    QtLocation.QGeoRouteRequest.RouteOptimizations,
    shortest=QtLocation.QGeoRouteRequest.ShortestRoute,
    fastest=QtLocation.QGeoRouteRequest.FastestRoute,
    most_economic=QtLocation.QGeoRouteRequest.MostEconomicRoute,
    most_scenic=QtLocation.QGeoRouteRequest.MostScenicRoute,
)

RouteOptimizationStr = Literal["shortest", "fastest", "most_economic", "most_scenic"]

SEGMENT_DETAIL = bidict(
    none=QtLocation.QGeoRouteRequest.NoSegmentData,
    basic=QtLocation.QGeoRouteRequest.BasicSegmentData,
)

SegmentDetailStr = Literal["none", "basic"]

TRAVEL_MODE = bidict(
    car=QtLocation.QGeoRouteRequest.CarTravel,
    pedestrian=QtLocation.QGeoRouteRequest.PedestrianTravel,
    bicycle=QtLocation.QGeoRouteRequest.BicycleTravel,
    public_transit=QtLocation.QGeoRouteRequest.PublicTransitTravel,
    truck=QtLocation.QGeoRouteRequest.TruckTravel,
)

TravelModeStr = Literal["car", "pedestrian", "bicycle", "public_transit", "truck"]


class GeoRouteRequest(QtLocation.QGeoRouteRequest):
    def get_waypoints(self) -> List[positioning.GeoCoordinate]:
        return [positioning.GeoCoordinate(wp) for wp in self.waypoints()]

    def get_exclude_areas(self) -> List[positioning.GeoRectangle]:
        return [positioning.GeoRectangle(wp) for wp in self.excludeAreas()]

    def get_departure_time(self) -> core.DateTime:
        return core.DateTime(self.departureTime())

    def set_feature_weight(self, feature: FeatureTypeStr, weight: FeatureWeightStr):
        """Set the feature weight.

        Args:
            feature: Feature type
            weight: Feature weight

        Raises:
            InvalidParamError: feature weight / type does not exist
        """
        if weight not in FEATURE_WEIGHTS:
            raise InvalidParamError(weight, FEATURE_WEIGHTS)
        if feature not in FEATURE_TYPES:
            raise InvalidParamError(feature, FEATURE_TYPES)
        self.setFeatureWeight(FEATURE_TYPES[feature], FEATURE_WEIGHTS[weight])

    def get_feature_weight(self, feature: FeatureTypeStr) -> FeatureWeightStr:
        """Return current feature weight.

        Returns:
            Feature weight
        """
        if feature not in FEATURE_TYPES:
            raise InvalidParamError(feature, FEATURE_TYPES)
        return FEATURE_WEIGHTS.inverse[self.featureWeight(FEATURE_TYPES[feature])]

    def set_route_optimization(self, optimization: RouteOptimizationStr):
        """Set the route optimization.

        Args:
            optimization: Route optimization

        Raises:
            InvalidParamError: route optimization does not exist
        """
        if optimization not in ROUTE_OPTIMIZATION:
            raise InvalidParamError(optimization, ROUTE_OPTIMIZATION)
        self.setRouteOptimization(ROUTE_OPTIMIZATION[optimization])

    def get_route_optimization(self) -> RouteOptimizationStr:
        """Return current route optimization.

        Returns:
            Route optimization
        """
        return ROUTE_OPTIMIZATION.inverse[self.routeOptimization()]

    def get_travel_modes(self) -> List[TravelModeStr]:
        return [k for k, v in TRAVEL_MODE.items() if v & self.travelModes()]

    def set_travel_modes(self, *mode: TravelModeStr):
        for item in mode:
            if item not in TRAVEL_MODE:
                raise InvalidParamError(item, TRAVEL_MODE)
        flags = helpers.merge_flags(mode, TRAVEL_MODE)
        self.setTravelModes(flags)

    def get_feature_types(self) -> List[FeatureTypeStr]:
        return [k for k, v in FEATURE_TYPES.items() for t in self.featureTypes() if v & t]


if __name__ == "__main__":
    request = GeoRouteRequest()



from __future__ import annotations

from typing import Literal, Tuple

from prettyqt import core
from prettyqt.qt import QtGui
from prettyqt.utils import bidict


TRANSFORMATION_TYPE = bidict(
    none=QtGui.QTransform.TxNone,
    translate=QtGui.QTransform.TxTranslate,
    scale=QtGui.QTransform.TxScale,
    rotate=QtGui.QTransform.TxRotate,
    shear=QtGui.QTransform.TxShear,
    project=QtGui.QTransform.TxProject,
)

TransformationTypeStr = Literal[
    "none", "translate", "scale", "rotate", "shear", "project"
]


class Transform(QtGui.QTransform):
    def __getstate__(self):
        return bytes(self)

    def __repr__(self):
        return (
            f"{type(self).__name__}({self.m11()}, {self.m12()}, {self.m13()}, "
            f"{self.m21()}, {self.m22()}, {self.m23()}, {self.m31()}, {self.m32()}, "
            f"{self.m33()})"
        )

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __setstate__(self, ba):
        core.DataStream.write_bytearray(ba, self)

    def __getitem__(self, value: Tuple[int, int]) -> float:
        if value[0] == 0:
            if value[1] == 0:
                return self.m11()
            elif value[1] == 1:
                return self.m12()
            elif value[1] == 2:
                return self.m13()
        elif value[0] == 1:
            if value[1] == 0:
                return self.m21()
            elif value[1] == 1:
                return self.m22()
            elif value[1] == 2:
                return self.m23()
        elif value[0] == 2:
            if value[1] == 0:
                return self.m31()
            elif value[1] == 1:
                return self.m32()
            elif value[1] == 2:
                return self.m33()
        raise ValueError(f"Wrong value {value}")

    def __bytes__(self):
        ba = core.DataStream.create_bytearray(self)
        return bytes(ba)

    @classmethod
    def clone_from(cls, transform: QtGui.QTransform) -> Transform:
        return cls(
            transform.m11(),
            transform.m12(),
            transform.m13(),
            transform.m21(),
            transform.m22(),
            transform.m23(),
            transform.m31(),
            transform.m32(),
            transform.m33(),
        )

    def get_type(self) -> TransformationTypeStr:
        return TRANSFORMATION_TYPE.inverse[self.type()]


if __name__ == "__main__":
    transform = Transform()
    print(transform[0, 0])
    print(repr(transform))



from __future__ import annotations

import datetime
from typing import Literal, Optional, Union

from prettyqt import widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict


SELECTION_MODE = bidict(
    none=QtWidgets.QCalendarWidget.NoSelection,
    single=QtWidgets.QCalendarWidget.SingleSelection,
)

SelectionModeStr = Literal["none", "single"]

VERTICAL_HEADER_FORMAT = bidict(
    none=QtWidgets.QCalendarWidget.NoVerticalHeader,
    week_numbers=QtWidgets.QCalendarWidget.ISOWeekNumbers,
)

VerticalHeaderFormatStr = Literal["none", "week_numbers"]

HORIZONTAL_HEADER_FORMAT = bidict(
    single_letter=QtWidgets.QCalendarWidget.SingleLetterDayNames,
    short=QtWidgets.QCalendarWidget.ShortDayNames,
    long=QtWidgets.QCalendarWidget.LongDayNames,
    none=QtWidgets.QCalendarWidget.NoHorizontalHeader,
)

HorizontalHeaderFormatStr = Literal["single_letter", "short", "long", "none"]


QtWidgets.QCalendarWidget.__bases__ = (widgets.Widget,)


class CalendarWidget(QtWidgets.QCalendarWidget):
    def serialize_fields(self):
        return dict(date=self.get_date())

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setSelectedDate(state["date"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def get_date(self) -> datetime.date:
        return self.selectedDate().toPython()  # type: ignore

    def get_value(self) -> datetime.date:
        return self.get_date()

    def set_value(self, value: Union[QtCore.QDate, datetime.date]):
        self.setSelectedDate(value)  # type: ignore

    def set_range(
        self,
        min_val: Union[QtCore.QDate, datetime.date],
        max_val: Union[QtCore.QDate, datetime.date],
    ):
        self.setMinimumDate(min_val)  # type: ignore
        self.setMaximumDate(max_val)  # type: ignore

    def set_selection_mode(self, mode: Optional[SelectionModeStr]):
        """Set selection mode for given calendar widget.

        Args:
            mode: selection mode to use

        Raises:
            InvalidParamError: mode does not exist
        """
        if mode is None:
            mode = "none"
        if mode not in SELECTION_MODE:
            raise InvalidParamError(mode, SELECTION_MODE)
        self.setSelectionMode(SELECTION_MODE[mode])

    def get_selection_mode(self) -> SelectionModeStr:
        """Return current selection mode.

        Returns:
            selection mode
        """
        return SELECTION_MODE.inverse[self.selectionMode()]


if __name__ == "__main__":
    app = widgets.app()
    w = CalendarWidget()
    w.show()
    app.main_loop()



from __future__ import annotations

from prettyqt.qt import QtQml
from prettyqt.utils import bidict


FLAGS = bidict(
    force_async_image_loading=QtQml.QQmlImageProviderBase.ForceAsynchronousImageLoading,
)


IMAGE_TYPE = bidict(
    image=QtQml.QQmlImageProviderBase.Image,
    pixmap=QtQml.QQmlImageProviderBase.Pixmap,
    texture=QtQml.QQmlImageProviderBase.Texture,
    image_response=QtQml.QQmlImageProviderBase.ImageResponse,
)


class QmlImageProviderBase(QtQml.QQmlImageProviderBase):
    pass



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QColumnView.__bases__ = (widgets.AbstractItemView,)


class ColumnView(QtWidgets.QColumnView):
    pass


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    dlg = QtWidgets.QMainWindow()
    status_bar = ColumnView()
    dlg.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets


# Remove dotted border on cell focus.  https://stackoverflow.com/a/55252650/3620725
class NoFocusDelegate(widgets.StyledItemDelegate):
    def paint(
        self,
        painter: QtGui.QPainter,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ):
        if option.state & widgets.Style.State_HasFocus:
            option.state = option.state ^ widgets.Style.State_HasFocus
        super().paint(painter, option, index)



from __future__ import annotations

from prettyqt import network
from prettyqt.qt import QtNetwork


QtNetwork.QTcpSocket.__bases__ = (network.AbstractSocket,)


class TcpSocket(QtNetwork.QTcpSocket):
    pass



from __future__ import annotations

import re
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

from prettyqt import core
from prettyqt.qt import QtCore
from prettyqt.utils import bidict


# IGNORECASE = QtCore.QRegularExpression.CaseInsensitiveOption
# MULTILINE = QtCore.QRegularExpression.MultilineOption
# DOTALL = QtCore.QRegularExpression.DotMatchesEverythingOption
# VERBOSE = QtCore.QRegularExpression.ExtendedPatternSyntaxOption

DONT_ESCAPE = {"!", '"', "%", "'", ",", "/", ":", ";", "<", "=", ">", "@", "`"}

IGNORECASE = re.IGNORECASE
MULTILINE = re.MULTILINE
DOTALL = re.DOTALL
VERBOSE = re.VERBOSE

MAP = bidict(
    {
        re.IGNORECASE: QtCore.QRegularExpression.CaseInsensitiveOption,
        re.MULTILINE: QtCore.QRegularExpression.MultilineOption,
        re.DOTALL: QtCore.QRegularExpression.DotMatchesEverythingOption,
        re.VERBOSE: QtCore.QRegularExpression.ExtendedPatternSyntaxOption,
    }
)


class Match(core.RegularExpressionMatch):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.string = None
        self.pos = None
        self.endpos = None

    def __repr__(self):
        return f"<re.Match object; span={self.span()}, match={self.groups()}>"

    def __getitem__(self, item: Union[str, int]) -> str:
        return self.captured(item)

    def group(self, *groups: Union[str, int]) -> Union[tuple, str]:
        if len(groups) > 1:
            return tuple(self.captured(i) for i in groups)
        if len(groups) == 0:
            return self.captured(0)
        return self.captured(groups[0])

    def groups(self, default=None) -> tuple:
        if self.lastindex is None:
            return tuple()
        return tuple(
            self.group(i) if i <= self.lastindex else default
            for i in range(self.re.captureCount())
        )

    def groupdict(self, default=None) -> Dict[str, Any]:
        if self.lastindex is None:
            return {}
        groups = [
            self.group(i) if i <= self.lastindex else default
            for i in range(self.re.captureCount())
        ]
        names = self.re.namedCaptureGroups()
        return {names[i]: groups[i] for i in range(self.re.captureCount())}

    def start(self, group: int = 0) -> int:
        return self.capturedStart(group)

    def end(self, group: int = 0) -> int:
        return self.capturedEnd(group)

    def span(self, group: int = 0) -> Tuple[int, int]:
        return (self.capturedStart(group), self.capturedEnd(group))

    @property
    def lastindex(self) -> Optional[int]:
        idx = self.lastCapturedIndex()
        return None if idx == -1 else idx

    @property
    def lastgroup(self) -> Optional[str]:
        if self.lastCapturedIndex() == -1:
            return None
        return self.re.namedCaptureGroups()[self.lastCapturedIndex()]

    @property
    def re(self) -> QtCore.QRegularExpression:
        return self.regularExpression()

    @property
    def partial(self) -> bool:
        return self.hasPartialMatch()


class Pattern(core.RegularExpression):
    def __init__(self, pattern: str = "", flags: int = 0):
        flag = QtCore.QRegularExpression.NoPatternOption
        for k, v in MAP.items():
            if k & flags:
                flag |= v
        super().__init__(pattern, flag)

    def match(  # type: ignore[override]
        self, string: str, pos: int = 0, endpos: Optional[int] = None
    ) -> Optional[Match]:
        match = super().match(string[:endpos], pos)
        return Match(match) if match.hasMatch() else None

    def fullmatch(
        self, string: str, pos: int = 0, endpos: Optional[int] = None
    ) -> Optional[Match]:
        if endpos:
            string = string[:endpos]
        match = super().match(string, pos)
        if match.hasMatch() and len(string) == match.end() - match.start():
            return Match(match)
        else:
            return None

    def finditer(
        self, string: str, pos: int = 0, endpos: Optional[int] = None
    ) -> Iterator[Match]:
        for match in self.globalMatch(string[:endpos], offset=pos):
            match.pos = pos
            match.endpos = endpos
            match.string = string
            yield match

    def findall(self, string: str, pos: int = 0, endpos: Optional[int] = None) -> list:
        matches = list(self.globalMatch(string[:endpos], offset=pos))
        return [m.groups() if len(m.groups()) > 1 else m.group(0) for m in matches]

    def subn(self, repl: Union[str, Callable], string: str, count: int = 0):
        result = string
        matches = self.global_match(string)
        matches = list(matches)
        if count > 0:
            matches = matches[:count]
        matches = list(reversed(matches))
        for m in matches:
            to_replace = repl if isinstance(repl, str) else repl(m)
            for j in range(self.groups):
                to_replace = to_replace.replace(fr"\g<{j}>", m.group(j))
            for k, v in self.groupindex.items():
                to_replace = to_replace.replace(fr"\g<{k}>", m.group(v))
            result = result[: m.start()] + to_replace + result[m.end() :]
        return (result, min(len(matches), count))

    def sub(self, repl: Union[str, Callable], string: str, count: int = 0):
        res = self.subn(repl, string, count)
        return res[0]

    def search(self, string: str, pos: int = 0, endpos: Optional[int] = None):
        match = super().match(string[:endpos], pos)
        return match if match.hasMatch() else None

    def split(self, string: str, maxsplit: int = 0):
        result = list()
        matches = self.global_match(string)
        matches = list(matches)
        prev_match = None
        num_split = min(maxsplit, len(matches)) if maxsplit > 0 else len(matches)
        if matches[0].start() == 0:
            result.append("")
        else:
            result.append(string[0 : matches[0].start()])
        for m in matches[:num_split]:
            if prev_match is not None:
                result.append(string[prev_match.end() : m.start()])
            for g in m.groups():
                result.append(g)
            prev_match = m
        if matches[num_split - 1].end() == len(string):
            result.append("")
        if maxsplit > 0:
            remainder_start = matches[maxsplit].end()
            result.append(string[remainder_start:])
        return result

    @property
    def groups(self) -> int:
        return self.captureCount()

    @property
    def groupindex(self) -> Dict[str, int]:
        return {k: i for i, k in enumerate(self.namedCaptureGroups()[1:], start=1)}

    @property
    def flags(self):
        return self.patternOptions()


def compile(pattern: str, flags: int = 0) -> Pattern:
    return Pattern(pattern, flags)


def search(pattern: str, string: str, flags: int = 0) -> Optional[core.RegularExpression]:
    compiled = compile(pattern, flags)
    match = compiled.search(string)
    return match


def match(pattern: str, string: str, flags: int = 0) -> Optional[core.RegularExpression]:
    compiled = compile(pattern, flags)
    return compiled.match(string)


def fullmatch(
    pattern: str, string: str, flags: int = 0
) -> Optional[core.RegularExpression]:
    compiled = compile(pattern, flags)
    return compiled.fullmatch(string)


# def split(pattern: str, string: str, maxsplit=0, flags=0) -> list:
#     compiled = compile(pattern, flags)
#     return compiled.split(string, maxsplit)


def findall(pattern: str, string: str, flags: int = 0) -> List[str]:
    compiled = compile(pattern, flags)
    return compiled.findall(string)


def finditer(pattern, string: str, flags: int = 0):
    compiled = compile(pattern, flags)
    return compiled.finditer(string)


def sub(pattern: str, repl: str, string: str, count: int = 0, flags: int = 0) -> str:
    compiled = compile(pattern, flags)
    return compiled.sub(repl, string, count)


def subn(
    pattern: str, repl: str, string: str, count: int = 0, flags: int = 0
) -> Tuple[str, int]:
    compiled = compile(pattern, flags)
    return compiled.subn(repl, string, count)


def escape(pattern: str):
    result = core.RegularExpression.escape(pattern)
    for i in DONT_ESCAPE:
        result.replace(i, r"\i")
    return result


if __name__ == "__main__":
    reg = compile("[0-9]+ [a-z] [0-9]+")



from __future__ import annotations

import os
import pathlib
from typing import List, Literal

from prettyqt import core
from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError, bidict, helpers


FILTERS = bidict(
    dirs=QtCore.QDir.Dirs,
    all_dirs=QtCore.QDir.AllDirs,
    files=QtCore.QDir.Files,
    drives=QtCore.QDir.Drives,
    no_sym_links=QtCore.QDir.NoSymLinks,
    no_dot_and_dotdot=QtCore.QDir.NoDotAndDotDot,
    no_dot=QtCore.QDir.NoDot,
    no_dotdot=QtCore.QDir.NoDotDot,
    all_entries=QtCore.QDir.AllEntries,
    readable=QtCore.QDir.Readable,
    writable=QtCore.QDir.Writable,
    executable=QtCore.QDir.Executable,
    modified=QtCore.QDir.Modified,
    hidden=QtCore.QDir.Hidden,
    system=QtCore.QDir.System,
    case_sensitive=QtCore.QDir.CaseSensitive,
)

FilterStr = Literal[
    "dirs",
    "all_dirs",
    "files",
    "drives",
    "no_sym_links",
    "no_dot_and_dotdot",
    "no_dot",
    "no_dotdot",
    "all_entries",
    "readable",
    "writable",
    "executable",
    "modified",
    "hidden",
    "system",
    "case_sensitive",
]


class Dir(QtCore.QDir):
    def __getattr__(self, attr: str):
        return getattr(self.to_path(), attr)

    def __repr__(self):
        return f"{type(self).__name__}({self.absolutePath()!r})"

    def __str__(self):
        return self.absolutePath()

    def __reduce__(self):
        return type(self), (self.absolutePath(),)

    def __truediv__(self, other: os.PathLike) -> pathlib.Path:
        return self.to_path() / os.fspath(other)

    def __fspath__(self) -> str:
        return self.absolutePath()

    def __bool__(self):
        return self.exists()

    def __abs__(self) -> str:
        return self.absolutePath()

    def to_path(self) -> pathlib.Path:
        return pathlib.Path(self.absolutePath())

    def set_filter(self, *filters: FilterStr):
        for item in filters:
            if item not in FILTERS:
                raise InvalidParamError(item, FILTERS)
        flags = helpers.merge_flags(filters, FILTERS)
        self.setFilter(flags)

    def get_filter(self) -> List[FilterStr]:
        return [k for k, v in FILTERS.items() if v & self.filter()]

    @classmethod
    def get_current(cls) -> Dir:
        return cls(cls.current())

    @classmethod
    def get_home(cls) -> Dir:
        return cls(cls.home())

    @classmethod
    def get_current_path(cls) -> pathlib.Path:
        return pathlib.Path(cls.currentPath())

    @classmethod
    def get_home_path(cls) -> pathlib.Path:
        return pathlib.Path(cls.homePath())

    @classmethod
    def get_drives(cls) -> List[core.FileInfo]:
        return [core.FileInfo(i) for i in cls.drives()]



from __future__ import annotations

from prettyqt import constants, core
from prettyqt.qt import QtGui
from prettyqt.utils import InvalidParamError


class FontMetricsF(QtGui.QFontMetricsF):
    def elided_text(
        self, text: str, mode: constants.ElideModeStr, width: float, flags=0
    ) -> str:
        if mode not in constants.ELIDE_MODE:
            raise InvalidParamError(mode, constants.ELIDE_MODE)
        return self.elidedText(text, constants.ELIDE_MODE[mode], width, flags)

    def get_bounding_rect(self, *args) -> core.RectF:
        return core.RectF(self.boundingRect(*args))

    def get_tight_bounding_rect(self, text: str) -> core.RectF:
        return core.RectF(self.tightBoundingRect(text))



from __future__ import annotations

from typing import Dict, List, Literal

from prettyqt import network
from prettyqt.qt import QtCore, QtNetwork
from prettyqt.utils import InvalidParamError, bidict, helpers


CAPABILITIES = bidict(
    tunneling=QtNetwork.QNetworkProxy.TunnelingCapability,
    listening=QtNetwork.QNetworkProxy.ListeningCapability,
    udp_tunneling=QtNetwork.QNetworkProxy.UdpTunnelingCapability,
    caching=QtNetwork.QNetworkProxy.CachingCapability,
    host_name_lookup=QtNetwork.QNetworkProxy.HostNameLookupCapability,
    sctp_tunneling=QtNetwork.QNetworkProxy.SctpTunnelingCapability,
    sctp_listening=QtNetwork.QNetworkProxy.SctpListeningCapability,
)

CapabilityStr = Literal[
    "tunneling",
    "listening",
    "udp_tunneling",
    "caching",
    "host_name_lookup",
    "sctp_tunneling",
    "sctp_listening",
]

PROXY_TYPES = bidict(
    none=QtNetwork.QNetworkProxy.NoProxy,
    default=QtNetwork.QNetworkProxy.DefaultProxy,
    socks5=QtNetwork.QNetworkProxy.Socks5Proxy,
    http=QtNetwork.QNetworkProxy.HttpProxy,
    http_caching=QtNetwork.QNetworkProxy.HttpCachingProxy,
    ftp_caching=QtNetwork.QNetworkProxy.FtpCachingProxy,
)

ProxyTypeStr = Literal[
    "none",
    "default",
    "socks5",
    "http",
    "http_caching",
    "ftp_caching",
]


class NetworkProxy(QtNetwork.QNetworkProxy):
    def get_capabilities(self) -> List[CapabilityStr]:
        return [k for k, v in CAPABILITIES.items() if v & self.capabilities()]

    def set_capabilities(self, *capability: CapabilityStr):
        for item in capability:
            if item not in CAPABILITIES:
                raise InvalidParamError(item, CAPABILITIES)
        flags = helpers.merge_flags(capability, CAPABILITIES)
        self.setCapabilities(flags)

    def get_header(self, name: network.networkrequest.KnownHeaderStr) -> str:
        if name not in network.networkrequest.KNOWN_HEADER:
            raise InvalidParamError(name, network.networkrequest.KNOWN_HEADER)
        return self.header(network.networkrequest.KNOWN_HEADER[name])

    def set_header(self, name: network.networkrequest.KnownHeaderStr, value: str):
        if name not in network.networkrequest.KNOWN_HEADER:
            raise InvalidParamError(name, network.networkrequest.KNOWN_HEADER)
        self.setHeader(network.networkrequest.KNOWN_HEADER[name], value)

    def get_headers(self) -> Dict[str, str]:
        return {
            bytes(h).decode(): bytes(self.rawHeader(h)).decode()
            for h in self.rawHeaderList()
        }

    def set_headers(self, headers: Dict[str, str]):
        for k, v in headers.items():
            self.setRawHeader(
                QtCore.QByteArray(k.encode()), QtCore.QByteArray(v.encode())
            )

    def set_type(self, typ: ProxyTypeStr):
        """Set proxy type.

        Args:
            typ: proxy type

        Raises:
            InvalidParamError: proxy type does not exist
        """
        if typ not in PROXY_TYPES:
            raise InvalidParamError(typ, PROXY_TYPES)
        self.setType(PROXY_TYPES[typ])

    def get_type(self) -> ProxyTypeStr:
        """Get the proxy type.

        Returns:
            type
        """
        return PROXY_TYPES.inverse[self.type()]



from __future__ import annotations

from typing import Literal, Union

from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError, bidict


ALGORITHM = bidict(
    md_4=QtCore.QCryptographicHash.Md4,
    md_5=QtCore.QCryptographicHash.Md5,
    sha_1=QtCore.QCryptographicHash.Sha1,
    sha_224=QtCore.QCryptographicHash.Sha224,
    sha_256=QtCore.QCryptographicHash.Sha256,
    sha_384=QtCore.QCryptographicHash.Sha384,
    sha_512=QtCore.QCryptographicHash.Sha512,
    sha3_224=QtCore.QCryptographicHash.Sha3_224,
    sha3_256=QtCore.QCryptographicHash.Sha3_256,
    sha3_384=QtCore.QCryptographicHash.Sha3_384,
    sha3_512=QtCore.QCryptographicHash.Sha3_512,
    keccak_224=QtCore.QCryptographicHash.Keccak_224,
    keccak_256=QtCore.QCryptographicHash.Keccak_256,
    keccak_384=QtCore.QCryptographicHash.Keccak_384,
    keccak_512=QtCore.QCryptographicHash.Keccak_512,
)

AlgorithmStr = Literal[
    "md_4",
    "md_5",
    "sha_1",
    "sha_224",
    "sha_256",
    "sha_384",
    "sha_512",
    "sha3_224",
    "sha3_256",
    "sha3_384",
    "sha3_512",
    "keccak_224",
    "keccak_256",
    "keccak_384",
    "keccak_512",
]


class CryptographicHash(QtCore.QCryptographicHash):
    def __init__(self, method: Union[QtCore.QCryptographicHash.Algorithm, AlgorithmStr]):
        if isinstance(method, QtCore.QCryptographicHash.Algorithm):
            arg = method
        else:
            arg = ALGORITHM[method]
        super().__init__(arg)

    # def __str__(self):
    #     return bytes(self.result()).decode()

    def __bytes__(self):
        return self.get_result()

    def get_result(self) -> bytes:
        return bytes(self.result())

    @staticmethod
    def get_hash_length(method: AlgorithmStr) -> int:
        if method not in ALGORITHM:
            raise InvalidParamError(method, ALGORITHM)
        return CryptographicHash.hashLength(ALGORITHM[method])



from __future__ import annotations

from prettyqt import charts
from prettyqt.qt.QtCharts import QtCharts


QtCharts.QScatterSeries.__bases__ = (charts.XYSeries,)


class ScatterSeries(QtCharts.QScatterSeries):
    pass



from __future__ import annotations

from prettyqt import gui
from prettyqt.qt import QtGui


QtGui.QPaintDeviceWindow.__bases__ = (gui.Window, gui.PaintDevice)


class PaintDeviceWindow(QtGui.QPaintDeviceWindow):
    pass


if __name__ == "__main__":
    window = PaintDeviceWindow()



from __future__ import annotations

from typing import Optional

from prettyqt import constants, widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError


QtWidgets.QTableView.__bases__ = (widgets.AbstractItemView,)


class TableView(QtWidgets.QTableView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        class_name = type(self).__name__
        self.set_id(class_name)
        self.setHorizontalHeader(widgets.HeaderView("horizontal", parent=self))
        self.setVerticalHeader(widgets.HeaderView("vertical", parent=self))
        self.setAlternatingRowColors(True)
        self.setWordWrap(False)

    def serialize_fields(self):
        return dict(
            corner_button_enabled=self.isCornerButtonEnabled(),
            grid_style=self.get_grid_style(),
            show_grid=self.showGrid(),
            sorting_enabled=self.isSortingEnabled(),
            word_wrap=self.wordWrap(),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setCornerButtonEnabled(state["corner_button_enabled"])
        self.set_grid_style(state["grid_style"])
        self.setShowGrid(state["show_grid"])
        self.setSortingEnabled(state["sorting_enabled"])
        self.setWordWrap(state["word_wrap"])

    @property
    def h_header(self):
        return self.horizontalHeader()

    @h_header.setter
    def h_header(self, header):
        self.setHorizontalHeader(header)

    @property
    def v_header(self):
        return self.verticalHeader()

    @v_header.setter
    def v_header(self, header):
        self.setVerticalHeader(header)

    def setup_list_style(self):
        self.set_selection_behaviour("rows")
        self.h_header.setStretchLastSection(True)
        self.v_header.set_resize_mode("fixed")
        self.v_header.set_default_section_size(28)

    def adapt_sizes(self):
        model = self.model()
        if model is not None and (model.rowCount() * model.columnCount()) < 1000:
            self.resizeColumnsToContents()
        else:
            self.h_header.resize_sections("interactive")

    def sort_by_column(self, column: Optional[int], ascending: bool = True):
        column = -1 if column is None else column
        order = constants.ASCENDING if ascending else constants.DESCENDING
        self.sortByColumn(column, order)

    def set_grid_style(self, style: constants.PenStyleStr):
        """Set grid style.

        Args:
            style: grid style to use

        Raises:
            InvalidParamError: invalid grid style
        """
        if style not in constants.PEN_STYLE:
            raise InvalidParamError(style, constants.PEN_STYLE)
        self.setGridStyle(constants.PEN_STYLE[style])

    def get_grid_style(self) -> constants.PenStyleStr:
        """Return grid style.

        Returns:
            grid style
        """
        return constants.PEN_STYLE.inverse[self.gridStyle()]


if __name__ == "__main__":
    app = widgets.app()
    widget = TableView()
    widget.set_model(widgets.FileSystemModel())
    widget.show()
    app.main_loop()



from __future__ import annotations

from typing import Iterator, Literal, Optional

from prettyqt import core, gui, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets
from prettyqt.utils import InvalidParamError, bidict


WIZARD_BUTTON = bidict(
    back=QtWidgets.QWizard.BackButton,
    next=QtWidgets.QWizard.NextButton,
    commit=QtWidgets.QWizard.CommitButton,
    finish=QtWidgets.QWizard.FinishButton,
    cancel=QtWidgets.QWizard.CancelButton,
    help=QtWidgets.QWizard.HelpButton,
    custom_1=QtWidgets.QWizard.CustomButton1,
    custom_2=QtWidgets.QWizard.CustomButton2,
    custom_3=QtWidgets.QWizard.CustomButton3,
)

WizardButtonStr = Literal[
    "back",
    "next",
    "commit",
    "finish",
    "cancel",
    "help",
    "custom_1",
    "custom_2",
    "custom_3",
]

WIZARD_OPTIONS = bidict(
    independent_pages=QtWidgets.QWizard.IndependentPages,
    ignore_subtitles=QtWidgets.QWizard.IgnoreSubTitles,
    extended_watermark_bitmap=QtWidgets.QWizard.ExtendedWatermarkPixmap,
    no_default_button=QtWidgets.QWizard.NoDefaultButton,
    no_back_button_on_start_page=QtWidgets.QWizard.NoBackButtonOnStartPage,
    no_back_button_on_last_page=QtWidgets.QWizard.NoBackButtonOnLastPage,
    disabled_back_button_on_last_page=QtWidgets.QWizard.DisabledBackButtonOnLastPage,
    next_button_on_last_page=QtWidgets.QWizard.HaveNextButtonOnLastPage,
    finish_button_on_early_pages=QtWidgets.QWizard.HaveFinishButtonOnEarlyPages,
    no_cancel_button=QtWidgets.QWizard.NoCancelButton,
    cancel_button_on_left=QtWidgets.QWizard.CancelButtonOnLeft,
    help_button=QtWidgets.QWizard.HaveHelpButton,
    help_button_on_right=QtWidgets.QWizard.HelpButtonOnRight,
    custom_button_1=QtWidgets.QWizard.HaveCustomButton1,
    custom_button_2=QtWidgets.QWizard.HaveCustomButton2,
    custom_button_3=QtWidgets.QWizard.HaveCustomButton3,
    no_cancel_button_on_last_page=QtWidgets.QWizard.NoCancelButtonOnLastPage,
)

WizardOptionStr = Literal[
    "independent_pages",
    "ignore_subtitles",
    "extended_watermark_bitmap",
    "no_default_button",
    "no_back_button_on_start_page",
    "no_back_button_on_last_page",
    "disabled_back_button_on_last_page",
    "next_button_on_last_page",
    "finish_button_on_early_pages",
    "no_cancel_button",
    "cancel_button_on_left",
    "help_button",
    "help_button_on_right",
    "custom_button_1",
    "custom_button_2",
    "custom_button_3",
    "no_cancel_button_on_last_page",
]

WIZARD_PIXMAP = bidict(
    watermark=QtWidgets.QWizard.WatermarkPixmap,
    logo=QtWidgets.QWizard.LogoPixmap,
    banner=QtWidgets.QWizard.BannerPixmap,
    background=QtWidgets.QWizard.BackgroundPixmap,
)

WizardPixmapStr = Literal["watermark", "logo", "banner", "background"]

WIZARD_STYLE = bidict(
    classic=QtWidgets.QWizard.ClassicStyle,
    modern=QtWidgets.QWizard.ModernStyle,
    mac=QtWidgets.QWizard.MacStyle,
    aero=QtWidgets.QWizard.AeroStyle,
)

WizardStyleStr = Literal["classic", "modern", "mac", "aero"]

TEXT_FORMATS = bidict(
    rich=QtCore.Qt.RichText, plain=QtCore.Qt.PlainText, auto=QtCore.Qt.AutoText
)

if core.VersionNumber.get_qt_version() >= (5, 14, 0):
    TEXT_FORMATS["markdown"] = QtCore.Qt.MarkdownText

QtWidgets.QWizard.__bases__ = (widgets.BaseDialog,)


class Wizard(QtWidgets.QWizard):
    def __getitem__(self, key: int) -> QtWidgets.QWizardPage:
        p = self.page(key)
        if p is None:
            raise KeyError(key)
        return p

    def __setitem__(self, key: int, value: QtWidgets.QWizardPage):
        return self.setPage(key, value)

    def __delitem__(self, key: int):
        if key not in self.pageIds():
            raise KeyError(key)
        return self.removePage(key)

    def __iter__(self) -> Iterator[QtWidgets.QWizardPage]:
        return iter(self.page(i) for i in self.pageIds())

    def __add__(self, other: QtWidgets.QWizardPage) -> Wizard:
        self.addPage(other)
        return self

    def add_widget_as_page(self, widget: QtWidgets.QWidget) -> None:
        page = widgets.WizardPage(self)
        layout = widgets.BoxLayout("vertical", self)
        layout += widget
        page.set_layout(layout)

    def set_title_format(self, fmt: str):
        """Set the title format.

        Allowed values are "rich", "plain", "auto", "markdown"

        Args:
            fmt: title format to use

        Raises:
            InvalidParamError: title format does not exist
        """
        if fmt not in TEXT_FORMATS:
            raise InvalidParamError(fmt, TEXT_FORMATS)
        self.setTitleFormat(TEXT_FORMATS[fmt])

    def get_title_format(self) -> str:
        """Return current title format.

        Possible values: "rich", "plain", "auto", "markdown"

        Returns:
            title format
        """
        return TEXT_FORMATS.inverse[self.titleFormat()]

    def set_subtitle_format(self, fmt: str):
        """Set the subtitle format.

        Allowed values are "rich", "plain", "auto", "markdown"

        Args:
            fmt: subtitle format to use

        Raises:
            InvalidParamError: subtitle format does not exist
        """
        if fmt not in TEXT_FORMATS:
            raise InvalidParamError(fmt, TEXT_FORMATS)
        self.setSubTitleFormat(TEXT_FORMATS[fmt])

    def get_subtitle_format(self) -> str:
        """Return current subtitle format.

        Possible values: "rich", "plain", "auto", "markdown"

        Returns:
            subtitle format
        """
        return TEXT_FORMATS.inverse[self.subTitleFormat()]

    def get_button(self, button_type: str) -> QtWidgets.QAbstractButton:
        if button_type not in WIZARD_BUTTON:
            raise InvalidParamError(button_type, WIZARD_BUTTON)
        return self.button(WIZARD_BUTTON[button_type])

    def set_button_text(self, button_type: WizardButtonStr, value: str):
        """Set text for given button type.

        Args:
            button_type: button to get text from
            value: text to set

        """
        if button_type not in WIZARD_BUTTON:
            raise InvalidParamError(button_type, WIZARD_BUTTON)
        self.setButtonText(WIZARD_BUTTON[button_type], value)

    def get_button_text(self, button_type: WizardButtonStr) -> str:
        """Return text for given button type.

        Args:
            button_type: button to get text from

        Returns:
            Button text
        """
        if button_type not in WIZARD_BUTTON:
            raise InvalidParamError(button_type, WIZARD_BUTTON)
        return self.buttonText(WIZARD_BUTTON[button_type])

    def set_pixmap(self, typ: WizardPixmapStr, pixmap: Optional[QtGui.QPixmap]):
        if typ not in WIZARD_PIXMAP:
            raise InvalidParamError(typ, WIZARD_PIXMAP)
        if pixmap is None:
            pixmap = QtGui.QPixmap()
        self.setPixmap(WIZARD_PIXMAP[typ], pixmap)

    def get_pixmap(self, typ: WizardPixmapStr) -> Optional[gui.Pixmap]:
        if typ not in WIZARD_PIXMAP:
            raise InvalidParamError(typ, WIZARD_PIXMAP)
        pix = gui.Pixmap(self.pixmap(WIZARD_PIXMAP[typ]))
        if pix.isNull():
            return None
        return pix

    def set_wizard_style(self, style: WizardStyleStr):
        """Set the wizard style.

        Args:
            style: wizard style

        Raises:
            InvalidParamError: wizard style does not exist
        """
        if style not in WIZARD_STYLE:
            raise InvalidParamError(style, WIZARD_STYLE)
        self.setWizardStyle(WIZARD_STYLE[style])

    def get_wizard_style(self) -> WizardStyleStr:
        """Return current wizard style.

        Returns:
            Wizard style
        """
        return WIZARD_STYLE.inverse[self.wizardStyle()]

    def set_option(self, option: WizardOptionStr, value: bool):
        """Set option to given value.

        Args:
            option: option to use
            value: value to set

        Raises:
            InvalidParamError: option does not exist
        """
        if option not in WIZARD_OPTIONS:
            raise InvalidParamError(option, WIZARD_OPTIONS)
        self.setOption(WIZARD_OPTIONS[option], value)

    def get_option(self, option: WizardOptionStr) -> bool:
        """Return the value assigned to option.

        Args:
            option: option to get

        Returns:
            option
        """
        if option not in WIZARD_OPTIONS:
            raise InvalidParamError(option, WIZARD_OPTIONS)
        return self.testOption(WIZARD_OPTIONS[option])


if __name__ == "__main__":
    app = widgets.app()
    dlg = Wizard()
    dlg.add_widget_as_page(widgets.RadioButton("test"))
    dlg.show()
    app.main_loop()



from dataclasses import dataclass
from typing import List, Optional

from prettyqt import constants, core, widgets
from prettyqt.qt import QtCore, QtWidgets


@dataclass
class ItemWrapper:
    item: QtWidgets.QWidgetItem
    position: int


class BorderLayout(widgets.Layout):
    West, North, South, East, Center = range(5)
    MinimumSize, SizeHint = range(2)

    def __init__(
        self,
        parent: Optional[QtWidgets.QWidget] = None,
        margin: int = 0,
        spacing: Optional[int] = None,
    ):
        super().__init__(parent)  # type: ignore
        self.set_margin(margin)
        self.setSpacing(spacing if spacing is not None else -1)
        self.items: List[ItemWrapper] = []

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item: QtWidgets.QWidgetItem):
        self.add_widgetitem(item, BorderLayout.West)

    def addWidget(self, widget: QtWidgets.QWidget, position: int = 0):
        self.add_widgetitem(widgets.WidgetItem(widget), position)

    def expandingDirections(self):
        return constants.HORIZONTAL | constants.VERTICAL

    def hasHeightForWidth(self) -> bool:
        return False

    def count(self) -> int:
        return len(self.items)

    def itemAt(self, index: int) -> Optional[QtWidgets.QWidgetItem]:  # type: ignore
        if index < len(self.items):
            return self.items[index].item
        return None

    def minimumSize(self):
        return self.calculate_size(BorderLayout.MinimumSize)

    def setGeometry(self, rect: QtCore.QRect):
        center = None
        east_width = 0
        west_width = 0
        north_height = 0
        south_height = 0

        super().setGeometry(rect)

        for wrapper in self.items:
            item = wrapper.item
            position = wrapper.position

            if position == BorderLayout.North:
                geom = core.Rect(
                    rect.x(), north_height, rect.width(), item.sizeHint().height()
                )
                item.setGeometry(geom)

                north_height += item.geometry().height() + self.spacing()

            elif position == BorderLayout.South:
                geom = core.Rect(
                    item.geometry().x(),
                    item.geometry().y(),
                    rect.width(),
                    item.sizeHint().height(),
                )
                item.setGeometry(geom)

                south_height += item.geometry().height() + self.spacing()
                geom = core.Rect(
                    rect.x(),
                    rect.y() + rect.height() - south_height + self.spacing(),
                    item.geometry().width(),
                    item.geometry().height(),
                )
                item.setGeometry(geom)

            elif position == BorderLayout.Center:
                center = wrapper

        center_height = rect.height() - north_height - south_height

        for wrapper in self.items:
            item = wrapper.item
            position = wrapper.position

            if position == BorderLayout.West:
                geom = core.Rect(
                    rect.x() + west_width,
                    north_height,
                    item.sizeHint().width(),
                    center_height,
                )
                item.setGeometry(geom)

                west_width += item.geometry().width() + self.spacing()

            elif position == BorderLayout.East:
                geom = core.Rect(
                    item.geometry().x(),
                    item.geometry().y(),
                    item.sizeHint().width(),
                    center_height,
                )
                item.setGeometry(geom)

                east_width += item.geometry().width() + self.spacing()

                geom = core.Rect(
                    rect.x() + rect.width() - east_width + self.spacing(),
                    north_height,
                    item.geometry().width(),
                    item.geometry().height(),
                )
                item.setGeometry(geom)

        if center:
            rect = core.Rect(
                west_width,
                north_height,
                rect.width() - east_width - west_width,
                center_height,
            )
            center.item.setGeometry(rect)

    def sizeHint(self) -> core.Size:
        return self.calculate_size(BorderLayout.SizeHint)

    def takeAt(self, index: int) -> Optional[QtWidgets.QWidgetItem]:  # type: ignore
        if 0 <= index < len(self.items):
            layout_struct = self.items.pop(index)
            return layout_struct.item

        return None

    def add_widgetitem(self, item: QtWidgets.QWidgetItem, position: int):
        self.items.append(ItemWrapper(item, position))

    def calculate_size(self, size_type: int) -> core.Size:
        total_size = core.Size()

        for wrapper in self.items:
            position = wrapper.position
            if size_type == BorderLayout.MinimumSize:
                item_size = wrapper.item.minimumSize()
            else:  # size_type == BorderLayout.SizeHint
                item_size = wrapper.item.sizeHint()

            if position in (BorderLayout.North, BorderLayout.South, BorderLayout.Center):
                total_size.setHeight(total_size.height() + item_size.height())

            if position in (BorderLayout.West, BorderLayout.East, BorderLayout.Center):
                total_size.setWidth(total_size.width() + item_size.width())

        return total_size


class Window(widgets.Widget):
    def __init__(self):
        super().__init__()

        central_widget = widgets.TextBrowser()
        central_widget.setPlainText("Central widget")

        layout = BorderLayout()
        layout.addWidget(central_widget, BorderLayout.Center)

        # Because BorderLayout doesn't call its super-class addWidget() it
        # doesn't take ownership of the widgets until setLayout() is called.
        # Therefore we keep a local reference to each label to prevent it being
        # garbage collected too soon.
        label_n = self.create_label("North")
        layout.addWidget(label_n, BorderLayout.North)

        label_w = self.create_label("West")
        layout.addWidget(label_w, BorderLayout.West)

        label_e1 = self.create_label("East 1")
        layout.addWidget(label_e1, BorderLayout.East)

        label_e2 = self.create_label("East 2")
        layout.addWidget(label_e2, BorderLayout.East)

        label_s = self.create_label("South")
        layout.addWidget(label_s, BorderLayout.South)

        self.setLayout(layout)

        self.setWindowTitle("Border Layout")

    @staticmethod
    def create_label(text: str):
        label = widgets.Label(text)
        label.setFrameStyle(widgets.Frame.Box | widgets.Frame.Raised)
        return label


if __name__ == "__main__":

    import sys

    app = widgets.app()
    window = Window()
    window.show()
    sys.exit(app.exec_())



from __future__ import annotations

from typing import Optional

from prettyqt import location
from prettyqt.qt import QtLocation


class PlaceCategory(QtLocation.QPlaceCategory):
    def __str__(self):
        return self.name()

    def __bool__(self):
        return not self.isEmpty()

    def get_icon(self) -> Optional[location.PlaceIcon]:
        icon = self.icon()
        if icon.isEmpty():
            return None
        return location.PlaceIcon(icon)

    def get_visibility(self) -> location.VisibilityStr:
        """Return the visibility of the place.

        Returns:
            Visibility
        """
        return location.VISIBILITY.inverse[self.visibility()]


if __name__ == "__main__":
    category = PlaceCategory()
    print(bool(category))



# from prettyqt.qt import QtMultimedia
# from prettyqt import multimedia

from __future__ import annotations


# QtMultimedia.QAbstractPlanarVideoBuffer.__bases__ = (multimedia.AbstractVideoBuffer,)

# class AbstractPlanarVideoBuffer(QtMultimedia.QAbstractPlanarVideoBuffer):
#     pass



from __future__ import annotations

from typing import Literal, Optional

from prettyqt import positioning
from prettyqt.qt import QtLocation
from prettyqt.utils import InvalidParamError, bidict


INSTRUCTION_DIRECTION = bidict(
    none=QtLocation.QGeoManeuver.NoDirection,
    forward=QtLocation.QGeoManeuver.DirectionForward,
    bear_right=QtLocation.QGeoManeuver.DirectionBearRight,
    light_right=QtLocation.QGeoManeuver.DirectionLightRight,
    right=QtLocation.QGeoManeuver.DirectionRight,
    hard_right=QtLocation.QGeoManeuver.DirectionHardRight,
    u_turn_right=QtLocation.QGeoManeuver.DirectionUTurnRight,
    u_turn_left=QtLocation.QGeoManeuver.DirectionUTurnLeft,
    hard_left=QtLocation.QGeoManeuver.DirectionHardLeft,
    left=QtLocation.QGeoManeuver.DirectionLeft,
    light_left=QtLocation.QGeoManeuver.DirectionLightLeft,
    bear_left=QtLocation.QGeoManeuver.DirectionBearLeft,
)

InstructionDirectionStr = Literal[
    "none",
    "forward",
    "bear_right",
    "light_right",
    "right",
    "hard_right",
    "u_turn_right",
    "u_turn_left",
    "hard_left",
    "left",
    "light_left",
    "bear_left",
]


class GeoManeuver(QtLocation.QGeoManeuver):
    def __bool__(self):
        return self.isValid()

    def __str__(self):
        return self.instructionText()

    def __setitem__(self, index: str, val):
        attrs = self.extendedAttributes()
        attrs[index] = val
        self.setExtendedAttributes(attrs)

    def __getitem__(self, index: str):
        attr = self.extendedAttributes()
        if index not in attr:
            raise KeyError(f"Key {index!r} does not exist.")
        return attr[index]

    def get_position(self) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.position())

    def get_waypoint(self) -> Optional[positioning.GeoCoordinate]:
        wp = self.waypoint()
        if not wp.isValid():
            return None
        return positioning.GeoCoordinate(wp)

    def set_direction(self, direction: InstructionDirectionStr):
        """Set the direction.

        Args:
            direction: Direction

        Raises:
            InvalidParamError: direction does not exist
        """
        if direction not in INSTRUCTION_DIRECTION:
            raise InvalidParamError(direction, INSTRUCTION_DIRECTION)
        self.setDirection(INSTRUCTION_DIRECTION[direction])

    def get_direction(self) -> InstructionDirectionStr:
        """Return current direction.

        Returns:
            Direction
        """
        return INSTRUCTION_DIRECTION.inverse[self.direction()]


if __name__ == "__main__":
    maneuver = GeoManeuver()
    print(bool(maneuver))
    maneuver.setExtendedAttributes(dict(a="a"))
    maneuver["test"] = "test"
    print(bool(maneuver))
    waypoint = maneuver.get_waypoint()
    print(bool(waypoint))
    print(maneuver.extendedAttributes())



from __future__ import annotations

from prettyqt import constants, core, gui, widgets
from prettyqt.qt import QtCore


class RenderLinkDelegate(widgets.StyledItemDelegate):
    def paint(self, painter, option, index):
        text = index.data()
        if not text:
            return

        painter.save()

        # I only wanted it for mouse over, but you'll probably want to remove
        # this condition
        if option.state and widgets.Style.State_MouseOver:
            font = option.font
            font.setUnderline(True)
            painter.setFont(font)
            painter.setPen(option.palette.link().color())
        painter.drawText(option.rect, QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter, text)
        painter.restore()

    def editorEvent(self, event, model, option, index):
        text = index.data()
        font = index.data(constants.FONT_ROLE)
        # alignment = index.data(constants.ALIGNMENT_ROLE)
        if font is None:
            font = widgets.Application.get_font()
        fm = gui.FontMetricsF(font)
        rect = fm.get_bounding_rect(
            core.RectF(option.rect),
            QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter,
            text,
        )
        if (
            event.type() == QtCore.QEvent.MouseButtonPress
            and event.button() == QtCore.Qt.LeftButton
            and event.localPos() in rect
        ):
            text = index.data()
            gui.DesktopServices.open_url(text)
            return True
        return False


if __name__ == "__main__":
    """Run the application."""
    from prettyqt import widgets

    app = widgets.app()

    # Create and populate the tableWidget
    table_widget = widgets.TableWidget(4, 2)
    table_widget.set_delegate(RenderLinkDelegate(), column=1)
    table_widget.set_selection_behaviour("rows")
    table_widget.setHorizontalHeaderLabels(["Title", "Progress"])

    data = [
        ["Mass in B-Minor", "http://www.google.de"],
        ["Three More Foxes", "http://www.google.com"],
        ["Sex Bomb", "http://www.google.de"],
        ["Barbie Girl", "http://www.google.de"],
    ]

    for i, r in enumerate(data):
        table_widget[i, 0] = widgets.TableWidgetItem(r[0])
        item = widgets.TableWidgetItem()
        item.setData(0, r[1])
        item.set_text_alignment("left")
        table_widget[i, 1] = item

    table_widget.resizeColumnsToContents()
    table_widget.resize(500, 300)
    table_widget.show()

    app.main_loop()



from __future__ import annotations

from typing import Union

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QSignalMapper.__bases__ = (core.Object,)


class SignalMapper(QtCore.QSignalMapper):
    def __getitem__(self, index: Union[int, str, QtCore.QObject]) -> QtCore.QObject:
        return self.mapping(index)

    def __setitem__(self, index: QtCore.QObject, value: Union[int, str, QtCore.QObject]):
        self.setMapping(index, value)


if __name__ == "__main__":
    mapper = SignalMapper()



from __future__ import annotations

from prettyqt import charts
from prettyqt.qt.QtCharts import QtCharts


QtCharts.QHorizontalBarSeries.__bases__ = (charts.AbstractBarSeries,)


class HorizontalBarSeries(QtCharts.QHorizontalBarSeries):
    pass



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtGui
from prettyqt.utils import bidict


WORD_WRAP_MODE = bidict(
    none=QtGui.QTextOption.NoWrap,
    word=QtGui.QTextOption.WordWrap,
    anywhere=QtGui.QTextOption.WrapAnywhere,
    boundary_or_anywhere=QtGui.QTextOption.WrapAtWordBoundaryOrAnywhere,
)

WordWrapModeStr = Literal["none", "word", "anywhere", "boundary_or_anywhere"]


class TextOption(QtGui.QTextOption):
    pass


if __name__ == "__main__":
    doc = TextOption()



from __future__ import annotations

from typing import Iterator, List, Optional, Tuple, Union

from prettyqt import constants, core, gui, iconprovider
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError, types


QtGui.QStandardItemModel.__bases__ = (core.AbstractItemModel,)


class StandardItemModel(QtGui.QStandardItemModel):
    def __getitem__(
        self, index: Union[int, Tuple[int, int], QtCore.QModelIndex]
    ) -> QtGui.QStandardItem:
        if isinstance(index, int):
            item = self.item(index)
        elif isinstance(index, tuple):
            item = self.item(*index)
        else:
            item = self.itemFromIndex(index)
        if item is None:
            raise KeyError(index)
        return item

    def __delitem__(self, index: Union[int, Tuple[int, int]]):
        if isinstance(index, int):
            item = self.takeRow(index)
        elif isinstance(index, tuple):
            item = self.takeItem(*index)
        if item is None:
            raise KeyError(index)
        return item

    def __iter__(self) -> Iterator[QtGui.QStandardItem]:
        return iter(self.get_children())

    def __getstate__(self):
        return dict(items=self.get_children())

    def __setstate__(self, state):
        for item in state["items"]:
            self.appendRow([item])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __add__(self, other: Union[str, QtGui.QStandardItem]) -> StandardItemModel:
        if isinstance(other, (QtGui.QStandardItem, str)):
            self.add(other)
            return self
        raise TypeError("wrong type for addition")

    def get_children(self) -> List[QtGui.QStandardItem]:
        return [self.item(index) for index in range(self.rowCount())]

    def add(self, *item: Union[str, QtGui.QStandardItem]):
        for i in item:
            if isinstance(i, str):
                i = gui.StandardItem(i)
            self.appendRow([i])

    def find_items(
        self, text: str, column: int = 0, mode: constants.MatchFlagStr = "exact"
    ) -> List[QtGui.QStandardItem]:
        if mode not in constants.MATCH_FLAGS:
            raise InvalidParamError(mode, constants.MATCH_FLAGS)
        return self.findItems(text, constants.MATCH_FLAGS[mode], column)

    def add_item(
        self,
        name: str = "",
        icon: types.IconType = None,
        data: Optional[dict] = None,
        foreground: Optional[QtGui.QBrush] = None,
        background: Optional[QtGui.QBrush] = None,
        font: Optional[QtGui.QFont] = None,
        selectable: bool = True,
        enabled: bool = True,
        editable: bool = False,
        status_tip: Optional[str] = None,
        tool_tip: Optional[str] = None,
        whats_this: Optional[str] = None,
        # text_alignment: Optional[str] = None,
        checkstate: Optional[constants.StateStr] = None,
        flags: Optional[QtCore.Qt.ItemFlags] = None,
        size_hint: Optional[QtCore.QSize] = None,
        is_user_type: bool = False,
    ) -> gui.StandardItem:
        item = gui.StandardItem(name)
        if icon is not None:
            icon = iconprovider.get_icon(icon)
            item.setIcon(icon)
        if data is not None:
            for k, v in data.items():
                item.setData(v, k)
        if foreground is not None:
            item.setForeground(foreground)
        if background is not None:
            item.setBackground(background)
        if font is not None:
            item.setFont(font)
        if flags is not None:
            item.setFlags(flags)
        if enabled:
            item.setEnabled(enabled)
        if editable:
            item.setEditable(editable)
        if selectable:
            item.setSelectable(selectable)
        if status_tip:
            item.setStatusTip(status_tip)
        if tool_tip:
            item.setToolTip(tool_tip)
        if whats_this:
            item.setWhatsThis(whats_this)
        if size_hint is not None:
            item.setSizeHint(size_hint)
        if checkstate is not None:
            item.set_checkstate(checkstate)
        self.appendRow([item])
        return item


if __name__ == "__main__":
    import pickle

    from prettyqt import widgets

    model = gui.StandardItemModel()
    model.add("test")
    app = widgets.app()
    w = widgets.ListView()
    w.set_model(model)
    model += gui.StandardItem("Item")
    for item in model:
        pass
    with open("data.pkl", "wb") as jar:
        pickle.dump(model, jar)
    with open("data.pkl", "rb") as jar:
        model = pickle.load(jar)
    model += gui.StandardItem("Item2")
    w.show()
    app.main_loop()
